# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/settings/rare.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_RareSettings(object):
    def setupUi(self, RareSettings):
        RareSettings.setObjectName("RareSettings")
        RareSettings.resize(623, 428)
        RareSettings.setWindowTitle("RareSettings")
        self.rare_layout = QtWidgets.QHBoxLayout(RareSettings)
        self.rare_layout.setObjectName("rare_layout")
        self.left_layout = QtWidgets.QVBoxLayout()
        self.left_layout.setObjectName("left_layout")
        self.interface_group = QtWidgets.QGroupBox(RareSettings)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.interface_group.sizePolicy().hasHeightForWidth())
        self.interface_group.setSizePolicy(sizePolicy)
        self.interface_group.setObjectName("interface_group")
        self.interface_layout = QtWidgets.QFormLayout(self.interface_group)
        self.interface_layout.setLabelAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.interface_layout.setObjectName("interface_layout")
        self.lang_label = QtWidgets.QLabel(self.interface_group)
        self.lang_label.setObjectName("lang_label")
        self.interface_layout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.lang_label)
        self.lang_select = QtWidgets.QComboBox(self.interface_group)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lang_select.sizePolicy().hasHeightForWidth())
        self.lang_select.setSizePolicy(sizePolicy)
        self.lang_select.setObjectName("lang_select")
        self.interface_layout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.lang_select)
        self.color_label = QtWidgets.QLabel(self.interface_group)
        self.color_label.setObjectName("color_label")
        self.interface_layout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.color_label)
        self.color_select = QtWidgets.QComboBox(self.interface_group)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.color_select.sizePolicy().hasHeightForWidth())
        self.color_select.setSizePolicy(sizePolicy)
        self.color_select.setObjectName("color_select")
        self.color_select.addItem("")
        self.interface_layout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.color_select)
        self.style_label = QtWidgets.QLabel(self.interface_group)
        self.style_label.setObjectName("style_label")
        self.interface_layout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.style_label)
        self.style_select = QtWidgets.QComboBox(self.interface_group)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.style_select.sizePolicy().hasHeightForWidth())
        self.style_select.setSizePolicy(sizePolicy)
        self.style_select.setObjectName("style_select")
        self.style_select.addItem("")
        self.interface_layout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.style_select)
        self.interface_info = QtWidgets.QLabel(self.interface_group)
        font = QtGui.QFont()
        font.setItalic(True)
        self.interface_info.setFont(font)
        self.interface_info.setWordWrap(True)
        self.interface_info.setObjectName("interface_info")
        self.interface_layout.setWidget(3, QtWidgets.QFormLayout.SpanningRole, self.interface_info)
        self.left_layout.addWidget(self.interface_group)
        self.settings_group = QtWidgets.QGroupBox(RareSettings)
        self.settings_group.setObjectName("settings_group")
        self.behavior_layout = QtWidgets.QGridLayout(self.settings_group)
        self.behavior_layout.setObjectName("behavior_layout")
        self.save_size = QtWidgets.QCheckBox(self.settings_group)
        self.save_size.setObjectName("save_size")
        self.behavior_layout.addWidget(self.save_size, 5, 0, 1, 1)
        self.notification = QtWidgets.QCheckBox(self.settings_group)
        self.notification.setObjectName("notification")
        self.behavior_layout.addWidget(self.notification, 4, 0, 1, 1)
        self.log_games = QtWidgets.QCheckBox(self.settings_group)
        self.log_games.setObjectName("log_games")
        self.behavior_layout.addWidget(self.log_games, 6, 0, 1, 1)
        self.sys_tray = QtWidgets.QCheckBox(self.settings_group)
        self.sys_tray.setObjectName("sys_tray")
        self.behavior_layout.addWidget(self.sys_tray, 0, 0, 1, 1)
        self.auto_update = QtWidgets.QCheckBox(self.settings_group)
        self.auto_update.setObjectName("auto_update")
        self.behavior_layout.addWidget(self.auto_update, 1, 0, 1, 1)
        self.confirm_start = QtWidgets.QCheckBox(self.settings_group)
        self.confirm_start.setObjectName("confirm_start")
        self.behavior_layout.addWidget(self.confirm_start, 2, 0, 1, 1)
        self.auto_sync_cloud = QtWidgets.QCheckBox(self.settings_group)
        self.auto_sync_cloud.setObjectName("auto_sync_cloud")
        self.behavior_layout.addWidget(self.auto_sync_cloud, 3, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.behavior_layout.addItem(spacerItem, 2, 1, 2, 1)
        self.left_layout.addWidget(self.settings_group)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.left_layout.addItem(spacerItem1)
        self.rare_layout.addLayout(self.left_layout)
        self.right_layout = QtWidgets.QVBoxLayout()
        self.right_layout.setObjectName("right_layout")
        self.log_dir_group = QtWidgets.QGroupBox(RareSettings)
        self.log_dir_group.setObjectName("log_dir_group")
        self.log_dir_layout = QtWidgets.QVBoxLayout(self.log_dir_group)
        self.log_dir_layout.setObjectName("log_dir_layout")
        self.log_dir_open_button = QtWidgets.QPushButton(self.log_dir_group)
        self.log_dir_open_button.setObjectName("log_dir_open_button")
        self.log_dir_layout.addWidget(self.log_dir_open_button)
        self.log_dir_clean_button = QtWidgets.QPushButton(self.log_dir_group)
        self.log_dir_clean_button.setObjectName("log_dir_clean_button")
        self.log_dir_layout.addWidget(self.log_dir_clean_button)
        self.log_dir_size_label = QtWidgets.QLabel(self.log_dir_group)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.log_dir_size_label.sizePolicy().hasHeightForWidth())
        self.log_dir_size_label.setSizePolicy(sizePolicy)
        self.log_dir_size_label.setText("")
        self.log_dir_size_label.setWordWrap(True)
        self.log_dir_size_label.setObjectName("log_dir_size_label")
        self.log_dir_layout.addWidget(self.log_dir_size_label)
        self.right_layout.addWidget(self.log_dir_group)
        self.groupBox = QtWidgets.QGroupBox(RareSettings)
        self.groupBox.setObjectName("groupBox")
        self.shortcuts_layout = QtWidgets.QVBoxLayout(self.groupBox)
        self.shortcuts_layout.setObjectName("shortcuts_layout")
        self.desktop_link_btn = QtWidgets.QPushButton(self.groupBox)
        self.desktop_link_btn.setObjectName("desktop_link_btn")
        self.shortcuts_layout.addWidget(self.desktop_link_btn)
        self.startmenu_link_btn = QtWidgets.QPushButton(self.groupBox)
        self.startmenu_link_btn.setObjectName("startmenu_link_btn")
        self.shortcuts_layout.addWidget(self.startmenu_link_btn)
        self.right_layout.addWidget(self.groupBox)
        spacerItem2 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.right_layout.addItem(spacerItem2)
        self.rare_layout.addLayout(self.right_layout)

        self.retranslateUi(RareSettings)

    def retranslateUi(self, RareSettings):
        _translate = QtCore.QCoreApplication.translate
        self.interface_group.setTitle(_translate("RareSettings", "Interface"))
        self.lang_label.setText(_translate("RareSettings", "Language"))
        self.color_label.setText(_translate("RareSettings", "Color scheme"))
        self.color_select.setItemText(0, _translate("RareSettings", "None"))
        self.style_label.setText(_translate("RareSettings", "Style sheet"))
        self.style_select.setItemText(0, _translate("RareSettings", "None"))
        self.interface_info.setText(_translate("RareSettings", "Restart Rare to apply changes."))
        self.settings_group.setTitle(_translate("RareSettings", "Behavior"))
        self.save_size.setText(_translate("RareSettings", "Restore window size on application startup"))
        self.notification.setText(_translate("RareSettings", "Show notifications when downloads complete"))
        self.log_games.setText(_translate("RareSettings", "Show console windows when launching games"))
        self.sys_tray.setText(_translate("RareSettings", "Exit to system tray"))
        self.auto_update.setText(_translate("RareSettings", "Queue game updates on application startup"))
        self.confirm_start.setText(_translate("RareSettings", "Confirm before launching games"))
        self.auto_sync_cloud.setText(_translate("RareSettings", "Automatically upload/download cloud saves"))
        self.log_dir_group.setTitle(_translate("RareSettings", "Logs"))
        self.log_dir_open_button.setText(_translate("RareSettings", "Open log folder"))
        self.log_dir_clean_button.setText(_translate("RareSettings", "Clean log folder"))
        self.groupBox.setTitle(_translate("RareSettings", "Shortcuts"))
        self.desktop_link_btn.setText(_translate("RareSettings", "Create on desktop"))
        self.startmenu_link_btn.setText(_translate("RareSettings", "Create in menu"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    RareSettings = QtWidgets.QWidget()
    ui = Ui_RareSettings()
    ui.setupUi(RareSettings)
    RareSettings.show()
    sys.exit(app.exec_())



from logging import getLogger

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QLayout, QMessageBox, QFrame
from legendary.core import LegendaryCore

from rare.shared import ArgumentsSingleton
from rare.ui.components.dialogs.login.landing_page import Ui_LandingPage
from rare.ui.components.dialogs.login.login_dialog import Ui_LoginDialog
from rare.utils.misc import icon
from rare.widgets.dialogs import BaseDialog
from rare.widgets.sliding_stack import SlidingStackedWidget
from .browser_login import BrowserLogin
from .import_login import ImportLogin

logger = getLogger("LoginDialog")


class LandingPage(QFrame):
    def __init__(self, parent=None):
        super(LandingPage, self).__init__(parent=parent)
        self.setFrameStyle(self.StyledPanel)
        self.ui = Ui_LandingPage()
        self.ui.setupUi(self)


class LoginDialog(BaseDialog):

    def __init__(self, core: LegendaryCore, parent=None):
        super(LoginDialog, self).__init__(parent=parent)
        self.setWindowFlags(
            Qt.Window
            | Qt.Dialog
            | Qt.CustomizeWindowHint
            | Qt.WindowSystemMenuHint
            | Qt.WindowTitleHint
            | Qt.WindowMinimizeButtonHint
            | Qt.WindowCloseButtonHint
            | Qt.MSWindowsFixedSizeDialogHint
        )

        self.ui = Ui_LoginDialog()
        self.ui.setupUi(self)

        self.logged_in: bool = False

        self.core = core
        self.args = ArgumentsSingleton()

        self.login_stack = SlidingStackedWidget(parent=self)
        self.login_stack.setMinimumWidth(480)
        self.ui.login_stack_layout.addWidget(self.login_stack)

        self.landing_page = LandingPage(self.login_stack)
        self.login_stack.insertWidget(0, self.landing_page)

        self.browser_page = BrowserLogin(self.core, self.login_stack)
        self.login_stack.insertWidget(1, self.browser_page)
        self.browser_page.success.connect(self.login_successful)
        self.browser_page.changed.connect(
            lambda: self.ui.next_button.setEnabled(self.browser_page.is_valid())
        )
        self.import_page = ImportLogin(self.core, self.login_stack)
        self.login_stack.insertWidget(2, self.import_page)
        self.import_page.success.connect(self.login_successful)
        self.import_page.changed.connect(lambda: self.ui.next_button.setEnabled(self.import_page.is_valid()))

        # # NOTE: The real problem is that the BrowserLogin page has a huge QLabel with word-wrapping enabled.
        # # That forces the whole form to vertically expand instead of horizontally. Since the form is not shown
        # # on the first page, the internal Qt calculation for the size of that form calculates it by expanding it
        # # vertically. Once the form becomes visible, the correct calculation takes place and that is why the
        # # dialog reduces in height. To avoid that, calculate the bounding size of all forms and set it as the
        # # minumum size
        # self.login_stack.setMinimumSize(
        #     self.landing_page.sizeHint().expandedTo(
        #         self.browser_page.sizeHint().expandedTo(self.import_page.sizeHint())
        #     )
        # )

        self.login_stack.setFixedHeight(
            max(
                self.landing_page.heightForWidth(self.login_stack.minimumWidth()),
                self.browser_page.heightForWidth(self.login_stack.minimumWidth()),
                self.import_page.heightForWidth(self.login_stack.minimumWidth()),
            )
        )

        self.ui.next_button.setEnabled(False)
        self.ui.back_button.setEnabled(False)

        self.landing_page.ui.login_browser_radio.clicked.connect(lambda: self.ui.next_button.setEnabled(True))
        self.landing_page.ui.login_browser_radio.clicked.connect(self.browser_radio_clicked)
        self.landing_page.ui.login_import_radio.clicked.connect(lambda: self.ui.next_button.setEnabled(True))
        self.landing_page.ui.login_import_radio.clicked.connect(self.import_radio_clicked)

        self.ui.exit_button.clicked.connect(self.reject)
        self.ui.back_button.clicked.connect(self.back_clicked)
        self.ui.next_button.clicked.connect(self.next_clicked)

        self.login_stack.setCurrentWidget(self.landing_page)

        self.ui.exit_button.setIcon(icon("fa.remove"))
        self.ui.back_button.setIcon(icon("fa.chevron-left"))
        self.ui.next_button.setIcon(icon("fa.chevron-right"))

        # lk: Set next as the default button only to stop closing the dialog when pressing enter
        self.ui.exit_button.setAutoDefault(False)
        self.ui.back_button.setAutoDefault(False)
        self.ui.next_button.setAutoDefault(True)

        self.ui.main_layout.setSizeConstraint(QLayout.SetFixedSize)

    def back_clicked(self):
        self.ui.back_button.setEnabled(False)
        self.ui.next_button.setEnabled(True)
        self.login_stack.slideInWidget(self.landing_page)

    def browser_radio_clicked(self):
        self.login_stack.slideInWidget(self.browser_page)
        self.ui.back_button.setEnabled(True)
        self.ui.next_button.setEnabled(False)

    def import_radio_clicked(self):
        self.login_stack.slideInWidget(self.import_page)
        self.ui.back_button.setEnabled(True)
        self.ui.next_button.setEnabled(self.import_page.is_valid())

    def next_clicked(self):
        if self.login_stack.currentWidget() is self.landing_page:
            if self.landing_page.ui.login_browser_radio.isChecked():
                self.browser_radio_clicked()
            if self.landing_page.ui.login_import_radio.isChecked():
                self.import_radio_clicked()
        elif self.login_stack.currentWidget() is self.browser_page:
            self.browser_page.do_login()
        elif self.login_stack.currentWidget() is self.import_page:
            self.import_page.do_login()

    def login(self):
        if self.args.test_start:
            self.reject()
        self.open()

    def login_successful(self):
        try:
            if self.core.login():
                self.logged_in = True
                self.accept()
            else:
                raise ValueError("Login failed.")
        except Exception as e:
            logger.error(str(e))
            self.core.lgd.invalidate_userdata()
            self.ui.next_button.setEnabled(False)
            self.logged_in = False
            QMessageBox.warning(None, self.tr("Login error"), str(e))




import os
from argparse import Namespace
from logging import getLogger

from PyQt5.QtCore import pyqtSignal, QObject

from rare.lgndr.cli import LegendaryCLI
from rare.lgndr.core import LegendaryCore
from rare.lgndr.glue.arguments import LgndrVerifyGameArgs
from rare.lgndr.glue.monkeys import LgndrIndirectStatus, verify_stdout_factory
from rare.models.game import RareGame
from .worker import QueueWorker, QueueWorkerInfo

logger = getLogger("VerifyWorker")


class VerifyWorker(QueueWorker):
    class Signals(QObject):
        progress = pyqtSignal(RareGame, int, int, float, float)
        result = pyqtSignal(RareGame, bool, int, int)
        error = pyqtSignal(RareGame, str)

    # num: int = 0
    # total: int = 1  # set default to 1 to avoid DivisionByZero before it is initialized

    def __init__(self, core: LegendaryCore, args: Namespace, rgame: RareGame):
        super(VerifyWorker, self).__init__()
        self.signals = VerifyWorker.Signals()
        self.core = core
        self.args = args
        self.rgame = rgame

    def __status_callback(self, num: int, total: int, percentage: float, speed: float):
        self.rgame.signals.progress.update.emit(num * 100 // total)
        self.signals.progress.emit(self.rgame, num, total, percentage, speed)

    def worker_info(self) -> QueueWorkerInfo:
        return QueueWorkerInfo(
            app_name=self.rgame.app_name, app_title=self.rgame.app_title, worker_type="Verify", state=self.state
        )

    def run_real(self):
        self.rgame.signals.progress.start.emit()

        cli = LegendaryCLI(self.core)
        status = LgndrIndirectStatus()
        args = LgndrVerifyGameArgs(
            app_name=self.rgame.app_name,
            indirect_status=status,
            verify_stdout=verify_stdout_factory(self.__status_callback)
        )

        # lk: first pass, verify with the current manifest
        repair_mode = False
        result = cli.verify_game(
            args, print_command=False, repair_mode=repair_mode, repair_online=not self.args.offline
        )
        if result is None:
            # lk: second pass with downloading the latest manifest
            # lk: this happens if the manifest was not found and repair_mode was not requested
            # lk: we already have checked if the directory exists before starting the worker
            try:
                # lk: this try-except block handles the exception caused by a missing manifest
                # lk: and is raised only in the case we are offline
                repair_mode = True
                result = cli.verify_game(
                    args, print_command=False, repair_mode=repair_mode, repair_online=not self.args.offline
                )
                if result is None:
                    raise ValueError
            except ValueError:
                self.rgame.signals.progress.finish.emit(True)
                self.signals.error.emit(self.rgame, status.message)
                return

        success = result is not None and not any(result)
        if success:
            # lk: if verification was successful we delete the repair file and run the clean procedure
            # lk: this could probably be cut down to what is relevant for this use-case and skip the `cli` call
            repair_file = os.path.join(self.core.lgd.get_tmp_path(), f"{self.rgame.app_name}.repair")
            cli.install_game_cleanup(
                game=self.rgame.game, igame=self.rgame.igame, repair_mode=True, repair_file=repair_file
            )
            self.rgame.needs_verification = False
            self.rgame.update_rgame()

        self.rgame.signals.progress.finish.emit(False)
        self.signals.result.emit(self.rgame, success, *result)



import os
from logging import getLogger

from PyQt5.QtCore import Qt, QSettings, QTimer, QSize, pyqtSignal, pyqtSlot
from PyQt5.QtGui import QCloseEvent, QCursor, QShowEvent
from PyQt5.QtWidgets import (
    QMainWindow,
    QApplication,
    QStatusBar,
    QScrollArea,
    QScroller,
    QComboBox,
    QMessageBox,
    QLabel,
    QWidget,
    QHBoxLayout,
)

from rare.components.tabs import MainTabWidget
from rare.components.tray_icon import TrayIcon
from rare.shared import RareCore
from rare.shared.workers.worker import QueueWorkerState
from rare.utils.paths import lock_file
from rare.widgets.elide_label import ElideLabel

logger = getLogger("MainWindow")


class MainWindow(QMainWindow):
    # int: exit code
    exit_app: pyqtSignal = pyqtSignal(int)

    def __init__(self, parent=None):
        self.__exit_code = 0
        self.__accept_close = False
        self._window_launched = False
        super(MainWindow, self).__init__(parent=parent)
        self.setAttribute(Qt.WA_DeleteOnClose, True)
        self.rcore = RareCore.instance()
        self.core = RareCore.instance().core()
        self.signals = RareCore.instance().signals()
        self.args = RareCore.instance().args()
        self.settings = QSettings()

        self.setWindowTitle("Rare - GUI for legendary")
        self.tab_widget = MainTabWidget(self)
        self.tab_widget.exit_app.connect(self.__on_exit_app)
        self.setCentralWidget(self.tab_widget)

        # Set up status bar stuff (jumping through a lot of hoops)
        self.status_bar = QStatusBar(self)
        self.setStatusBar(self.status_bar)

        self.active_label = QLabel(self.tr("Active:"), self.status_bar)
        # lk: set top and botton margins to accommodate border for scroll area labels
        self.active_label.setContentsMargins(5, 1, 0, 1)
        self.status_bar.addWidget(self.active_label)

        self.active_container = QWidget(self.status_bar)
        active_layout = QHBoxLayout(self.active_container)
        active_layout.setContentsMargins(0, 0, 0, 0)
        active_layout.setSizeConstraint(QHBoxLayout.SetFixedSize)
        self.status_bar.addWidget(self.active_container, stretch=0)

        self.queued_label = QLabel(self.tr("Queued:"), self.status_bar)
        # lk: set top and botton margins to accommodate border for scroll area labels
        self.queued_label.setContentsMargins(5, 1, 0, 1)
        self.status_bar.addPermanentWidget(self.queued_label)

        self.queued_scroll = QScrollArea(self.status_bar)
        self.queued_scroll.setFrameStyle(QScrollArea.NoFrame)
        self.queued_scroll.setWidgetResizable(True)
        self.queued_scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.queued_scroll.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.queued_scroll.setFixedHeight(self.queued_label.sizeHint().height())
        self.status_bar.addPermanentWidget(self.queued_scroll, stretch=1)

        self.queued_container = QWidget(self.queued_scroll)
        queued_layout = QHBoxLayout(self.queued_container)
        queued_layout.setContentsMargins(0, 0, 0, 0)
        queued_layout.setSizeConstraint(QHBoxLayout.SetFixedSize)

        self.active_label.setVisible(False)
        self.active_container.setVisible(False)
        self.queued_label.setVisible(False)
        self.queued_scroll.setVisible(False)

        self.signals.application.update_statusbar.connect(self.update_statusbar)

        # self.status_timer = QTimer(self)
        # self.status_timer.timeout.connect(self.update_statusbar)
        # self.status_timer.setInterval(5000)
        # self.status_timer.start()

        width, height = 1280, 720
        if self.settings.value("save_size", False, bool):
            width, height = self.settings.value("window_size", (width, height), tuple)

        self.resize(width, height)

        if not self.args.offline:
            try:
                from rare.utils.rpc import DiscordRPC

                self.rpc = DiscordRPC()
            except ModuleNotFoundError:
                logger.warning("Discord RPC module not found")

        self.timer = QTimer()
        self.timer.setInterval(1000)
        self.timer.timeout.connect(self.timer_finished)
        self.timer.start()

        self.tray_icon: TrayIcon = TrayIcon(self)
        self.tray_icon.exit_app.connect(self.__on_exit_app)
        self.tray_icon.show_app.connect(self.show)
        self.tray_icon.activated.connect(lambda r: self.toggle() if r == self.tray_icon.DoubleClick else None)

        # enable kinetic scrolling
        for scroll_area in self.findChildren(QScrollArea):
            if not scroll_area.property("no_kinetic_scroll"):
                QScroller.grabGesture(scroll_area.viewport(), QScroller.LeftMouseButtonGesture)

            # fix scrolling
            for combo_box in scroll_area.findChildren(QComboBox):
                combo_box.wheelEvent = lambda e: e.ignore()

    def center_window(self):
        # get the margins of the decorated window
        margins = self.windowHandle().frameMargins()
        # get the screen the cursor is on
        current_screen = QApplication.screenAt(QCursor.pos())
        if not current_screen:
            current_screen = QApplication.primaryScreen()
        # get the available screen geometry (excludes panels/docks)
        screen_rect = current_screen.availableGeometry()
        decor_width = margins.left() + margins.right()
        decor_height = margins.top() + margins.bottom()
        window_size = QSize(self.width(), self.height()).boundedTo(
            screen_rect.size() - QSize(decor_width, decor_height)
        )

        self.resize(window_size)
        self.move(screen_rect.center() - self.rect().adjusted(0, 0, decor_width, decor_height).center())

    @pyqtSlot()
    def show(self) -> None:
        super(MainWindow, self).show()
        if not self._window_launched:
            self.center_window()
        self._window_launched = True

    def hide(self) -> None:
        if self.settings.value("save_size", False, bool):
            size = self.size().width(), self.size().height()
            self.settings.setValue("window_size", size)
        super(MainWindow, self).hide()

    def toggle(self):
        if self.isHidden():
            self.show()
        else:
            self.hide()

    @pyqtSlot()
    def update_statusbar(self):
        self.active_label.setVisible(False)
        self.active_container.setVisible(False)
        self.queued_label.setVisible(False)
        self.queued_scroll.setVisible(False)
        for label in self.active_container.findChildren(QLabel, options=Qt.FindDirectChildrenOnly):
            self.active_container.layout().removeWidget(label)
            label.deleteLater()
        for label in self.queued_container.findChildren(QLabel, options=Qt.FindDirectChildrenOnly):
            self.queued_container.layout().removeWidget(label)
            label.deleteLater()
        for info in self.rcore.queue_info():
            label = ElideLabel(info.app_title)
            label.setObjectName("QueueWorkerLabel")
            label.setToolTip(f"<b>{info.worker_type}</b>: {info.app_title}")
            label.setProperty("workerType", info.worker_type)
            label.setFixedWidth(150)
            label.setContentsMargins(3, 0, 3, 0)
            if info.state == QueueWorkerState.ACTIVE:
                self.active_container.layout().addWidget(label)
                self.active_label.setVisible(True)
                self.active_container.setVisible(True)
            elif info.state == QueueWorkerState.QUEUED:
                self.queued_container.layout().addWidget(label)
                self.queued_label.setVisible(True)
                self.queued_scroll.setVisible(True)

    def timer_finished(self):
        file_path = lock_file()
        if os.path.exists(file_path):
            file = open(file_path, "r")
            action = file.read()
            file.close()
            if action.startswith("show"):
                self.show()
            os.remove(file_path)
        self.timer.start()

    @pyqtSlot()
    @pyqtSlot(int)
    def __on_exit_app(self, exit_code=0) -> None:
        self.__exit_code = exit_code
        self.close()

    def close(self) -> bool:
        self.__accept_close = True
        return super(MainWindow, self).close()

    def closeEvent(self, e: QCloseEvent) -> None:
        # lk: `accept_close` is set to `True` by the `close()` method, overrides exiting to tray in `closeEvent()`
        # lk: ensures exiting instead of hiding when `close()` is called programmatically
        if not self.__accept_close:
            if self.settings.value("sys_tray", True, bool):
                self.hide()
                e.ignore()
                return

        # FIXME: Move this to RareCore once the download manager is implemented
        if self.rcore.queue_threadpool.activeThreadCount():
            reply = QMessageBox.question(
                self,
                self.tr("Quit {}?").format(QApplication.applicationName()),
                self.tr(
                    "There are currently running operations. "
                    "Rare cannot exit until they are completed.\n\n"
                    "Do you want to clear the queue?"
                ),
                buttons=(QMessageBox.Yes | QMessageBox.No),
                defaultButton=QMessageBox.No,
            )
            if reply == QMessageBox.Yes:
                self.rcore.queue_threadpool.clear()
                for qw in self.rcore.queued_workers():
                    self.rcore.dequeue_worker(qw)
                self.update_statusbar()
            e.ignore()
            return
        elif self.tab_widget.downloads_tab.is_download_active:
            reply = QMessageBox.question(
                self,
                self.tr("Quit {}?").format(QApplication.applicationName()),
                self.tr(
                    "There is an active download. "
                    "Quitting Rare now will stop the download.\n\n"
                    "Are you sure you want to quit?"
                ),
                buttons=(QMessageBox.Yes | QMessageBox.No),
                defaultButton=QMessageBox.No,
            )
            if reply == QMessageBox.Yes:
                self.tab_widget.downloads_tab.stop_download(omit_queue=True)
            else:
                e.ignore()
                return
        # FIXME: End of FIXME
        self.timer.stop()
        self.tray_icon.deleteLater()
        self.hide()
        self.exit_app.emit(self.__exit_code)
        super(MainWindow, self).closeEvent(e)




from collections import deque
from enum import IntEnum
from logging import getLogger
from typing import Optional, Deque

from PyQt5.QtCore import pyqtSignal, pyqtSlot, Qt
from PyQt5.QtWidgets import (
    QWidget,
    QGroupBox,
    QVBoxLayout,
    QLabel,
    QSizePolicy,
)
from legendary.models.game import Game, InstalledGame

from rare.components.tabs.downloads.widgets import QueueWidget, UpdateWidget
from rare.models.install import InstallOptionsModel, InstallQueueItemModel
from rare.utils.misc import widget_object_name

logger = getLogger("QueueGroup")


class UpdateGroup(QGroupBox):
    update_count = pyqtSignal(int)
    enqueue = pyqtSignal(InstallOptionsModel)

    def __init__(self, parent=None):
        super(UpdateGroup, self).__init__(parent=parent)
        self.setObjectName(type(self).__name__)
        self.setTitle(self.tr("Updates"))
        self.__text = QLabel(self.tr("No updates available"))
        self.__text.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

        # lk: For findChildren to work, the update's layout has to be in a widget
        self.__container = QWidget(self)
        self.__container.setVisible(False)
        container_layout = QVBoxLayout(self.__container)
        container_layout.setContentsMargins(0, 0, 0, 0)

        layout = QVBoxLayout(self)
        layout.addWidget(self.__text)
        layout.addWidget(self.__container)

    def __find_widget(self, app_name: str) -> Optional[UpdateWidget]:
        return self.__container.findChild(UpdateWidget, name=widget_object_name(UpdateWidget, app_name))

    def count(self) -> int:
        return len(self.__container.findChildren(UpdateWidget, options=Qt.FindDirectChildrenOnly))

    def contains(self, app_name: str) -> bool:
        return self.__find_widget(app_name) is not None

    def __update_group(self):
        count = self.count()
        self.__text.setVisible(not count)
        self.__container.setVisible(bool(count))
        self.update_count.emit(count)

    def append(self, game: Game, igame: InstalledGame):
        self.__text.setVisible(False)
        self.__container.setVisible(True)
        widget: UpdateWidget = self.__find_widget(game.app_name)
        if widget is not None:
            self.__container.layout().removeWidget(widget)
            widget.deleteLater()
        widget = UpdateWidget(game, igame, parent=self.__container)
        widget.destroyed.connect(self.__update_group)
        widget.enqueue.connect(self.enqueue)
        self.__container.layout().addWidget(widget)

    def remove(self, app_name: str):
        widget: UpdateWidget = self.__find_widget(app_name)
        self.__container.layout().removeWidget(widget)
        widget.deleteLater()

    def set_widget_enabled(self, app_name: str, enabled: bool):
        widget: UpdateWidget = self.__find_widget(app_name)
        widget.set_enabled(enabled)

    def get_widget_version(self, app_name: str) -> str:
        widget: UpdateWidget = self.__find_widget(app_name)
        return widget.version()


class QueueGroup(QGroupBox):
    update_count = pyqtSignal(int)
    removed = pyqtSignal(str)
    force = pyqtSignal(InstallQueueItemModel)

    def __init__(self, parent=None):
        super(QueueGroup, self).__init__(parent=parent)
        self.setObjectName(type(self).__name__)
        self.setTitle(self.tr("Queue"))
        self.__text = QLabel(self.tr("No downloads in queue"), self)
        self.__text.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

        # lk: For findChildren to work, the queue's layout has to be in a widget
        self.__container = QWidget(self)
        self.__container.setVisible(False)
        container_layout = QVBoxLayout(self.__container)
        container_layout.setContentsMargins(0, 0, 0, 0)

        layout = QVBoxLayout(self)
        layout.addWidget(self.__text)
        layout.addWidget(self.__container)

        self.__queue: Deque[str] = deque()

    def __find_widget(self, app_name: str) -> Optional[QueueWidget]:
        return self.__container.findChild(QueueWidget, name=widget_object_name(QueueWidget, app_name))

    def count(self) -> int:
        return len(self.__queue)

    def contains(self, app_name: str) -> bool:
        if app_name in self.__queue:
            return self.__find_widget(app_name) is not None
        else:
            return False

    def __update_group(self):
        count = self.count()
        self.__text.setVisible(not count)
        self.__container.setVisible(bool(count))
        self.update_count.emit(count)

    def __create_widget(self, item: InstallQueueItemModel, old_igame: InstalledGame) -> QueueWidget:
        widget: QueueWidget = QueueWidget(item, old_igame, parent=self.__container)
        widget.toggle_arrows(self.__queue.index(item.options.app_name), len(self.__queue))
        widget.destroyed.connect(self.__update_group)
        widget.destroyed.connect(self.__update_arrows)
        widget.remove.connect(self.remove)
        widget.force.connect(self.__on_force)
        widget.move_up.connect(self.__on_move_up)
        widget.move_down.connect(self.__on_move_down)
        return widget

    def push_front(self, item: InstallQueueItemModel, old_igame: InstalledGame):
        self.__text.setVisible(False)
        self.__container.setVisible(True)
        self.__queue.appendleft(item.options.app_name)
        widget = self.__create_widget(item, old_igame)
        self.__container.layout().insertWidget(0, widget)
        if self.count() > 1:
            app_name = self.__queue[1]
            other: QueueWidget = self.__find_widget(app_name)
            other.toggle_arrows(1, len(self.__queue))

    def push_back(self, item: InstallQueueItemModel, old_igame: InstalledGame):
        self.__text.setVisible(False)
        self.__container.setVisible(True)
        self.__queue.append(item.download.game.app_name)
        widget = self.__create_widget(item, old_igame)
        self.__container.layout().addWidget(widget)
        if self.count() > 1:
            app_name = self.__queue[-2]
            other: QueueWidget = self.__find_widget(app_name)
            other.toggle_arrows(len(self.__queue) - 2, len(self.__queue))

    def pop_front(self) -> InstallQueueItemModel:
        app_name = self.__queue.popleft()
        widget: QueueWidget = self.__find_widget(app_name)
        item = widget.item
        widget.deleteLater()
        return item

    def __update_arrows(self):
        """
        Check the first, second, last and second to last widgets in the list
        and update their arrows
        :return: None
        """
        for idx in [0, 1]:
            if self.count() > idx:
                app_name = self.__queue[idx]
                widget: QueueWidget = self.__find_widget(app_name)
                widget.toggle_arrows(idx, len(self.__queue))
        for idx in [1, 2]:
            if self.count() > idx:
                app_name = self.__queue[-idx]
                widget: QueueWidget = self.__find_widget(app_name)
                widget.toggle_arrows(len(self.__queue) - idx, len(self.__queue))

    def __remove(self, app_name: str):
        self.__queue.remove(app_name)
        widget: QueueWidget = self.__find_widget(app_name)
        self.__container.layout().removeWidget(widget)
        widget.deleteLater()

    @pyqtSlot(str)
    def remove(self, app_name: str):
        self.__remove(app_name)
        self.removed.emit(app_name)

    @pyqtSlot(InstallQueueItemModel)
    def __on_force(self, item: InstallQueueItemModel):
        self.__remove(item.options.app_name)
        self.force.emit(item)

    class MoveDirection(IntEnum):
        UP = -1
        DOWN = 1

    def __move(self, app_name: str, direction: MoveDirection):
        """
        Moved the widget for `app_name` up or down in the queue and the container
        :param app_name: The app_name associated with the widget
        :param direction: -1 to move up, +1 to move down
        :return: None
        """
        index = self.__queue.index(app_name)
        self.__queue.remove(app_name)
        self.__queue.insert(index + int(direction), app_name)
        widget: QueueWidget = self.__find_widget(app_name)
        self.__container.layout().insertWidget(index + int(direction), widget)
        self.__update_arrows()

    @pyqtSlot(str)
    def __on_move_up(self, app_name: str):
        self.__move(app_name, QueueGroup.MoveDirection.UP)

    @pyqtSlot(str)
    def __on_move_down(self, app_name: str):
        self.__move(app_name, QueueGroup.MoveDirection.DOWN)



import os
import platform
from logging import getLogger
from typing import Tuple

from PyQt5.QtCore import QThreadPool, QSettings
from PyQt5.QtWidgets import (
    QWidget,
    QFileDialog,
    QLabel,
    QPushButton,
    QSizePolicy,
    QMessageBox,
    QGroupBox,
    QVBoxLayout,
    QSpacerItem,
)
from legendary.models.game import SaveGameStatus

from rare.models.game import RareGame
from rare.shared import RareCore
from rare.shared.workers.wine_resolver import WineResolver
from rare.ui.components.tabs.games.game_info.cloud_widget import Ui_CloudWidget
from rare.ui.components.tabs.games.game_info.sync_widget import Ui_SyncWidget
from rare.utils.misc import icon
from rare.widgets.indicator_edit import PathEdit, IndicatorReasonsCommon
from rare.widgets.loading_widget import LoadingWidget
from rare.widgets.side_tab import SideTabContents

logger = getLogger("CloudWidget")


class CloudSaves(QWidget, SideTabContents):
    def __init__(self, parent=None):
        super(CloudSaves, self).__init__(parent=parent)

        self.sync_widget = QWidget(self)
        self.sync_ui = Ui_SyncWidget()
        self.sync_ui.setupUi(self.sync_widget)

        self.info_label = QLabel(self.tr("<b>This game doesn't support cloud saves</b>"))

        self.rcore = RareCore.instance()
        self.core = RareCore.instance().core()
        self.settings = QSettings()

        self.sync_ui.icon_local.setPixmap(icon("mdi.harddisk", "fa.desktop").pixmap(128, 128))
        self.sync_ui.icon_remote.setPixmap(icon("mdi.cloud-outline", "ei.cloud").pixmap(128, 128))

        self.sync_ui.upload_button.clicked.connect(self.upload)
        self.sync_ui.download_button.clicked.connect(self.download)

        self.loading_widget = LoadingWidget(parent=self.sync_widget)
        self.loading_widget.setVisible(False)

        self.rgame: RareGame = None

        self.cloud_widget = QGroupBox(self)
        self.cloud_ui = Ui_CloudWidget()
        self.cloud_ui.setupUi(self.cloud_widget)

        self.cloud_save_path_edit = PathEdit(
            "",
            file_mode=QFileDialog.DirectoryOnly,
            placeholder=self.tr('Use "Calculate path" or "Browse" ...'),
            edit_func=self.edit_save_path,
            save_func=self.save_save_path,
        )
        self.cloud_ui.cloud_layout.addRow(QLabel(self.tr("Save path")), self.cloud_save_path_edit)

        self.compute_save_path_button = QPushButton(icon("fa.magic"), self.tr("Calculate path"))
        self.compute_save_path_button.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Fixed)
        self.compute_save_path_button.clicked.connect(self.compute_save_path)
        self.cloud_ui.cloud_layout.addRow(None, self.compute_save_path_button)

        self.cloud_ui.cloud_sync.stateChanged.connect(
            lambda: self.settings.setValue(
                f"{self.rgame.app_name}/auto_sync_cloud", self.cloud_ui.cloud_sync.isChecked()
            )
        )

        layout = QVBoxLayout(self)
        layout.addWidget(self.sync_widget)
        layout.addWidget(self.cloud_widget)
        layout.addWidget(self.info_label)
        layout.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Fixed, QSizePolicy.Expanding))

    def edit_save_path(self, text: str) -> Tuple[bool, str, int]:
        if platform.system() != "Windows":
            if os.path.exists(text):
                return True, text, IndicatorReasonsCommon.VALID
            else:
                return False, text, IndicatorReasonsCommon.DIR_NOT_EXISTS
        return True, text, IndicatorReasonsCommon.VALID

    def save_save_path(self, text: str):
        if text != self.rgame.save_path:
            self.rgame.save_path = text

    def upload(self):
        self.sync_widget.setEnabled(False)
        self.loading_widget.setVisible(True)
        self.rgame.upload_saves()

    def download(self):
        self.sync_widget.setEnabled(False)
        self.loading_widget.setVisible(True)
        self.rgame.download_saves()

    def compute_save_path(self):
        if self.rgame.is_installed and self.rgame.game.supports_cloud_saves:
            try:
                new_path = self.core.get_save_path(self.rgame.app_name)
                if platform.system() != "Windows" and not os.path.exists(new_path):
                    raise ValueError(f'Path "{new_path}" does not exist.')
            except Exception as e:
                logger.warning(str(e))
                resolver = WineResolver(self.core, self.rgame.raw_save_path, self.rgame.app_name)
                if not resolver.wine_env.get("WINEPREFIX"):
                    del resolver
                    self.cloud_save_path_edit.setText("")
                    QMessageBox.warning(self, "Warning", "No wine prefix selected. Please set it in settings")
                    return
                self.cloud_save_path_edit.setText(self.tr("Loading..."))
                self.cloud_save_path_edit.setDisabled(True)
                self.compute_save_path_button.setDisabled(True)

                app_name = self.rgame.app_name
                resolver.signals.result_ready.connect(lambda x: self.wine_resolver_finished(x, app_name))
                QThreadPool.globalInstance().start(resolver)
                return
            else:
                self.cloud_save_path_edit.setText(new_path)

    def wine_resolver_finished(self, path, app_name):
        logger.info(f"Wine resolver finished for {app_name}. Computed save path: {path}")
        if app_name == self.rgame.app_name:
            self.cloud_save_path_edit.setDisabled(False)
            self.compute_save_path_button.setDisabled(False)
            if path and not os.path.exists(path):
                try:
                    os.makedirs(path)
                except PermissionError:
                    self.cloud_save_path_edit.setText("")
                    QMessageBox.warning(
                        self,
                        self.tr("Error - {}").format(self.rgame.app_title),
                        self.tr(
                            "Error while calculating path for <b>{}</b>. Insufficient permissions to create <b>{}</b>"
                        ).format(self.rgame.app_title, path),
                    )
                    return
            if not path:
                self.cloud_save_path_edit.setText("")
                return
            self.cloud_save_path_edit.setText(path)
        elif path:
            self.rcore.get_game(app_name).save_path = path

    def __update_widget(self):
        supports_saves = self.rgame.igame is not None and (
                self.rgame.game.supports_cloud_saves or self.rgame.game.supports_mac_cloud_saves
        )

        self.sync_widget.setEnabled(
            bool(supports_saves and self.rgame.save_path))  # and not self.rgame.is_save_up_to_date))

        self.cloud_widget.setEnabled(supports_saves)
        self.info_label.setVisible(not supports_saves)
        if not supports_saves:
            self.sync_ui.date_info_local.setText("None")
            self.sync_ui.age_label_local.setText("None")
            self.sync_ui.date_info_remote.setText("None")
            self.sync_ui.age_label_remote.setText("None")
            self.cloud_ui.cloud_sync.setChecked(False)
            self.cloud_save_path_edit.setText("")
            return

        status, (dt_local, dt_remote) = self.rgame.save_game_state

        self.sync_ui.date_info_local.setText(
            dt_local.strftime("%A, %d. %B %Y %X") if dt_local and self.rgame.save_path else "None"
        )
        self.sync_ui.date_info_remote.setText(
            dt_remote.strftime("%A, %d. %B %Y %X") if dt_remote and self.rgame.save_path else "None"
        )

        newer = self.tr("Newer")
        self.sync_ui.age_label_local.setText(
            f"<b>{newer}</b>" if status == SaveGameStatus.LOCAL_NEWER else " "
        )
        self.sync_ui.age_label_remote.setText(
            f"<b>{newer}</b>" if status == SaveGameStatus.REMOTE_NEWER else " "
        )

        button_disabled = self.rgame.state in [RareGame.State.RUNNING, RareGame.State.SYNCING]
        self.sync_widget.setDisabled(button_disabled)
        if self.rgame.state == RareGame.State.SYNCING:
            self.loading_widget.start()
        else:
            self.loading_widget.stop()

        self.sync_ui.upload_button.setDisabled(not dt_local)
        self.sync_ui.download_button.setDisabled(not dt_remote)

        self.cloud_ui.cloud_sync.blockSignals(True)
        self.cloud_ui.cloud_sync.setChecked(self.rgame.auto_sync_saves)
        self.cloud_ui.cloud_sync.blockSignals(False)

        self.cloud_save_path_edit.setText(self.rgame.save_path if self.rgame.save_path else "")
        if platform.system() == "Windows" and not self.rgame.save_path:
            self.compute_save_path()

    def update_game(self, rgame: RareGame):
        if self.rgame:
            self.rgame.signals.widget.update.disconnect(self.__update_widget)

        self.rgame = rgame

        self.set_title.emit(rgame.app_title)
        rgame.signals.widget.update.connect(self.__update_widget)

        self.__update_widget()



# https://github.com/pycontribs/tendo/blob/master/tendo/singleton.py


import logging
import os
import sys
import tempfile

logger = logging.getLogger("tendo.singleton")

if sys.platform != "win32":
    import fcntl


class SingleInstanceException(BaseException):
    pass


class SingleInstance(object):
    """Class that can be instantiated only once per machine.

    If you want to prevent your script from running in parallel just instantiate SingleInstance() class. If is there another instance already running it will throw a `SingleInstanceException`.


    This option is very useful if you have scripts executed by crontab at small amounts of time.

    Remember that this works by creating a lock file with a filename based on the full path to the script file.

    Providing a flavor_id will augment the filename with the provided flavor_id, allowing you to create multiple singleton instances from the same file. This is particularly useful if you want specific functions to have their own singleton instances.
    """

    def __init__(self, flavor_id="", lockfile=""):
        self.initialized = False
        if lockfile:
            self.lockfile = lockfile
        else:
            basename = (
                os.path.splitext(os.path.abspath(sys.argv[0]))[0]
                .replace("/", "-")
                .replace(":", "")
                .replace("\\", "-")
                + "-%s" % flavor_id
                + ".lock"
            )
            self.lockfile = os.path.normpath(f"{tempfile.gettempdir()}/{basename}")

        logger.debug(f"SingleInstance lockfile: {self.lockfile}")
        if sys.platform == "win32":
            try:
                # file already exists, we try to remove (in case previous
                # execution was interrupted)
                if os.path.exists(self.lockfile):
                    os.unlink(self.lockfile)
                self.fd = os.open(self.lockfile, os.O_CREAT | os.O_EXCL | os.O_RDWR)
            except OSError:
                type, e, tb = sys.exc_info()
                if e.errno == 13:
                    logger.error("Another instance is already running, quitting.")
                    raise SingleInstanceException()
                print(e.errno)
                raise
        else:  # non Windows
            self.fp = open(self.lockfile, "w")
            self.fp.flush()
            try:
                fcntl.lockf(self.fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
            except IOError:
                logger.warning("Another instance is already running, quitting.")
                raise SingleInstanceException()
        self.initialized = True

    def __del__(self):
        if not self.initialized:
            return
        try:
            if sys.platform == "win32":
                if hasattr(self, "fd"):
                    os.close(self.fd)
                    os.unlink(self.lockfile)
            else:
                fcntl.lockf(self.fp, fcntl.LOCK_UN)
                # os.close(self.fp)
                if os.path.isfile(self.lockfile):
                    os.unlink(self.lockfile)
        except Exception as e:
            if logger:
                logger.warning(e)
            else:
                print("Unloggable error: %s" % e)
            sys.exit(-1)



import platform as pf
import re
from logging import getLogger
from typing import Tuple, List

from PyQt5.QtCore import QObject, pyqtSignal, QThreadPool, QSettings
from PyQt5.QtWidgets import QSizePolicy, QWidget, QFileDialog, QMessageBox

from rare.shared import LegendaryCoreSingleton
from rare.shared.workers.worker import Worker
from rare.ui.components.tabs.settings.legendary import Ui_LegendarySettings
from rare.utils.misc import format_size
from rare.widgets.indicator_edit import PathEdit, IndicatorLineEdit, IndicatorReasonsCommon

logger = getLogger("LegendarySettings")


class RefreshGameMetaWorker(Worker):
    class Signals(QObject):
        finished = pyqtSignal()

    def __init__(self, platforms: List[str], include_unreal: bool):
        super(RefreshGameMetaWorker, self).__init__()
        self.signals = RefreshGameMetaWorker.Signals()
        self.core = LegendaryCoreSingleton()
        if platforms:
            self.platforms = platforms
        else:
            self.platforms = ["Windows"]
        self.skip_ue = not include_unreal

    def run_real(self) -> None:
        for platform in self.platforms:
            self.core.get_game_and_dlc_list(
                True, platform=platform, force_refresh=True, skip_ue=self.skip_ue
            )
        self.signals.finished.emit()


class LegendarySettings(QWidget, Ui_LegendarySettings):
    def __init__(self, parent=None):
        super(LegendarySettings, self).__init__(parent=parent)
        self.setupUi(self)
        self.settings = QSettings(self)

        self.core = LegendaryCoreSingleton()

        # Platform specific installation directory for macOS games
        if pf.system() == "Darwin":
            self.mac_install_dir = PathEdit(
                self.core.get_default_install_dir("Mac"),
                placeholder=self.tr("Default installation folder for macOS games"),
                file_mode=QFileDialog.DirectoryOnly,
                save_func=self.__mac_path_save,
            )
            self.install_dir_layout.addWidget(self.mac_install_dir)

        # Platform-independent installation directory
        self.install_dir = PathEdit(
            self.core.get_default_install_dir(),
            placeholder=self.tr("Default installation folder for Windows games"),
            file_mode=QFileDialog.DirectoryOnly,
            save_func=self.__win_path_save,
        )
        self.install_dir_layout.addWidget(self.install_dir)

        # Max Workers
        max_workers = self.core.lgd.config["Legendary"].getint(
            "max_workers", fallback=0
        )
        self.max_worker_spin.setValue(max_workers)
        self.max_worker_spin.valueChanged.connect(self.max_worker_save)
        # Max memory
        max_memory = self.core.lgd.config["Legendary"].getint("max_memory", fallback=0)
        self.max_memory_spin.setValue(max_memory)
        self.max_memory_spin.valueChanged.connect(self.max_memory_save)
        # Preferred CDN
        preferred_cdn = self.core.lgd.config["Legendary"].get(
            "preferred_cdn", fallback=""
        )
        self.preferred_cdn_line.setText(preferred_cdn)
        self.preferred_cdn_line.textChanged.connect(self.preferred_cdn_save)
        # Disable HTTPS
        disable_https = self.core.lgd.config["Legendary"].getboolean(
            "disable_https", fallback=False
        )
        self.disable_https_check.setChecked(disable_https)
        self.disable_https_check.stateChanged.connect(self.disable_https_save)

        # Cleanup
        self.clean_button.clicked.connect(lambda: self.cleanup(False))
        self.clean_keep_manifests_button.clicked.connect(lambda: self.cleanup(True))

        self.locale_edit = IndicatorLineEdit(
            f"{self.core.language_code}-{self.core.country_code}",
            edit_func=self.locale_edit_cb,
            save_func=self.locale_save_cb,
            horiz_policy=QSizePolicy.Minimum,
            parent=self,
        )
        self.locale_layout.addWidget(self.locale_edit)

        self.fetch_win32_check.setChecked(self.settings.value("win32_meta", False, bool))
        self.fetch_win32_check.stateChanged.connect(
            lambda: self.settings.setValue("win32_meta", self.fetch_win32_check.isChecked())
        )

        self.fetch_macos_check.setChecked(self.settings.value("macos_meta", pf.system() == "Darwin", bool))
        self.fetch_macos_check.stateChanged.connect(
            lambda: self.settings.setValue("macos_meta", self.fetch_macos_check.isChecked())
        )
        self.fetch_macos_check.setDisabled(pf.system() == "Darwin")

        self.fetch_unreal_check.setChecked(self.settings.value("unreal_meta", False, bool))
        self.fetch_unreal_check.stateChanged.connect(
            lambda: self.settings.setValue("unreal_meta", self.fetch_unreal_check.isChecked())
        )

        self.refresh_metadata_button.clicked.connect(self.refresh_metadata)
        # FIXME: Disable the button for now because it interferes with RareCore
        self.refresh_metadata_button.setEnabled(False)
        self.refresh_metadata_button.setVisible(False)

    def refresh_metadata(self):
        self.refresh_metadata_button.setDisabled(True)
        platforms = []
        if self.fetch_win32_check.isChecked():
            platforms.append("Win32")
        if self.fetch_macos_check.isChecked():
            platforms.append("Mac")
        worker = RefreshGameMetaWorker(platforms, self.fetch_unreal_check.isChecked())
        worker.signals.finished.connect(lambda: self.refresh_metadata_button.setDisabled(False))
        QThreadPool.globalInstance().start(worker)

    @staticmethod
    def locale_edit_cb(text: str) -> Tuple[bool, str, int]:
        if text:
            if re.match("^[a-zA-Z]{2,3}[-_][a-zA-Z]{2,3}$", text):
                language, country = text.replace("_", "-").split("-")
                text = "-".join([language.lower(), country.upper()])
            if bool(re.match("^[a-z]{2,3}-[A-Z]{2,3}$", text)):
                return True, text, IndicatorReasonsCommon.VALID
            else:
                return False, text, IndicatorReasonsCommon.WRONG_FORMAT
        else:
            return True, text, IndicatorReasonsCommon.VALID

    def locale_save_cb(self, text: str):
        if text:
            self.core.egs.language_code, self.core.egs.country_code = text.split("-")
            self.core.lgd.config.set("Legendary", "locale", text)
        else:
            if self.core.lgd.config.has_option("Legendary", "locale"):
                self.core.lgd.config.remove_option("Legendary", "locale")
        self.core.lgd.save_config()

    def __mac_path_save(self, text: str) -> None:
        self.__path_save(text, "mac_install_dir")

    def __win_path_save(self, text: str) -> None:
        self.__path_save(text, "install_dir")
        if pf.system() != "Darwin":
            self.__mac_path_save(text)

    def __path_save(self, text: str, option: str = "Windows"):
        self.core.lgd.config["Legendary"][option] = text
        if not text and option in self.core.lgd.config["Legendary"].keys():
            self.core.lgd.config["Legendary"].pop(option)
        else:
            logger.debug(f"Set %s option in config to %s", option, text)
        self.core.lgd.save_config()

    def max_worker_save(self, workers: str):
        if workers := int(workers):
            self.core.lgd.config.set("Legendary", "max_workers", str(workers))
        else:
            self.core.lgd.config.remove_option("Legendary", "max_workers")
        self.core.lgd.save_config()

    def max_memory_save(self, memory: str):
        if memory := int(memory):
            self.core.lgd.config.set("Legendary", "max_memory", str(memory))
        else:
            self.core.lgd.config.remove_option("Legendary", "max_memory")
        self.core.lgd.save_config()

    def preferred_cdn_save(self, cdn: str):
        if cdn:
            self.core.lgd.config.set("Legendary", "preferred_cdn", cdn.strip())
        else:
            self.core.lgd.config.remove_option("Legendary", "preferred_cdn")
        self.core.lgd.save_config()

    def disable_https_save(self, checked: int):
        self.core.lgd.config.set(
            "Legendary", "disable_https", str(bool(checked)).lower()
        )
        self.core.lgd.save_config()

    def cleanup(self, keep_manifests: bool):
        before = self.core.lgd.get_dir_size()
        logger.debug("Removing app metadata...")
        app_names = set(g.app_name for g in self.core.get_assets(update_assets=False))
        self.core.lgd.clean_metadata(app_names)

        if not keep_manifests:
            logger.debug("Removing manifests...")
            installed = [
                (ig.app_name, ig.version, ig.platform) for ig in self.core.get_installed_list()
            ]
            installed.extend(
                (ig.app_name, ig.version, ig.platform) for ig in self.core.get_installed_dlc_list()
            )
            self.core.lgd.clean_manifests(installed)

        logger.debug("Removing tmp data")
        self.core.lgd.clean_tmp_data()

        after = self.core.lgd.get_dir_size()
        logger.info(
            f"Cleanup complete! Removed {(before - after) / 1024 / 1024:.02f} MiB."
        )
        if (before - after) > 0:
            QMessageBox.information(
                self,
                "Cleanup",
                self.tr("Cleanup complete! Successfully removed {}").format(
                    format_size(before - after)
                ),
            )
        else:
            QMessageBox.information(self, "Cleanup", "Nothing to clean")



import os
import platform
from abc import abstractmethod
from logging import getLogger
from typing import Tuple, Iterable, List, Union

from PyQt5.QtCore import Qt, QThreadPool, QRunnable, pyqtSlot, pyqtSignal
from PyQt5.QtGui import QShowEvent
from PyQt5.QtWidgets import QGroupBox, QListWidgetItem, QFileDialog, QMessageBox, QFrame, QFormLayout
from legendary.models.egl import EGLManifest
from legendary.models.game import InstalledGame

from rare.lgndr.glue.exception import LgndrException
from rare.models.pathspec import PathSpec
from rare.shared import RareCore
from rare.shared.workers.wine_resolver import WineResolver
from rare.ui.components.tabs.games.integrations.egl_sync_group import Ui_EGLSyncGroup
from rare.ui.components.tabs.games.integrations.egl_sync_list_group import Ui_EGLSyncListGroup
from rare.widgets.elide_label import ElideLabel
from rare.widgets.indicator_edit import PathEdit, IndicatorReasonsCommon

logger = getLogger("EGLSync")


class EGLSyncGroup(QGroupBox):
    def __init__(self, parent=None):
        super(EGLSyncGroup, self).__init__(parent=parent)
        self.ui = Ui_EGLSyncGroup()
        self.ui.setupUi(self)
        self.core = RareCore.instance().core()

        self.egl_path_edit = PathEdit(
            path=self.core.egl.programdata_path,
            placeholder=self.tr(
                "Path to the Wine prefix where EGL is installed, or the Manifests folder"
            ),
            file_mode=QFileDialog.DirectoryOnly,
            edit_func=self.egl_path_edit_edit_cb,
            save_func=self.egl_path_edit_save_cb,
            parent=self,
        )
        self.ui.egl_sync_layout.setWidget(
            self.ui.egl_sync_layout.getWidgetPosition(self.ui.egl_path_edit_label)[0],
            QFormLayout.FieldRole, self.egl_path_edit
        )

        self.egl_path_info = ElideLabel(parent=self)
        self.egl_path_info.setProperty("infoLabel", 1)
        self.ui.egl_sync_layout.setWidget(
            self.ui.egl_sync_layout.getWidgetPosition(self.ui.egl_path_info_label)[0],
            QFormLayout.FieldRole, self.egl_path_info
        )

        if platform.system() == "Windows":
            self.ui.egl_path_edit_label.setEnabled(False)
            self.egl_path_edit.setEnabled(False)
            self.ui.egl_path_info_label.setEnabled(False)
            self.egl_path_info.setEnabled(False)
        else:
            self.egl_path_edit.textChanged.connect(self.egl_path_changed)
            if self.core.egl.programdata_path:
                self.ui.egl_path_info_label.setEnabled(True)
                self.egl_path_info.setEnabled(True)

        self.ui.egl_sync_check.setChecked(self.core.egl_sync_enabled)
        self.ui.egl_sync_check.stateChanged.connect(self.egl_sync_changed)

        # lk: Temporarily disable automatic sync with EGL
        self.ui.egl_sync_check_label.setHidden(True)
        self.ui.egl_sync_check.setHidden(True)

        self.import_list = EGLSyncImportGroup(parent=self)
        self.ui.import_export_layout.addWidget(self.import_list)
        self.export_list = EGLSyncExportGroup(parent=self)
        self.ui.import_export_layout.addWidget(self.export_list)

        # self.egl_watcher = QFileSystemWatcher([self.egl_path_edit.text()], self)
        # self.egl_watcher.directoryChanged.connect(self.update_lists)

    def showEvent(self, a0: QShowEvent) -> None:
        if a0.spontaneous():
            return super().showEvent(a0)
        if not self.core.egl.programdata_path:
            self.__run_wine_resolver()
        self.update_lists()
        super().showEvent(a0)

    def __run_wine_resolver(self):
        self.egl_path_info.setText(self.tr("Updating..."))
        wine_resolver = WineResolver(
            self.core,
            PathSpec.egl_programdata,
            "default"
        )
        wine_resolver.signals.result_ready.connect(self.__on_wine_resolver_result)
        QThreadPool.globalInstance().start(wine_resolver)

    def __on_wine_resolver_result(self, path):
        self.egl_path_info.setText(path)
        if not path:
            self.egl_path_info.setText(
                self.tr(
                    "Default Wine prefix is unset, or path does not exist. "
                    "Create it or configure it in Settings -> Linux."
                )
            )
        elif not os.path.exists(path):
            self.egl_path_info.setText(
                self.tr(
                    "Default Wine prefix is set but EGL manifests path does not exist. "
                    "Your configured default Wine prefix might not be where EGL is installed."
                )
            )
        else:
            self.egl_path_edit.setText(path)

    @staticmethod
    def egl_path_edit_edit_cb(path) -> Tuple[bool, str, int]:
        if not path:
            return True, path, IndicatorReasonsCommon.VALID
        if os.path.exists(os.path.join(path, "system.reg")) and os.path.exists(
                os.path.join(path, "dosdevices/c:")
        ):
            # path is a wine prefix
            path = os.path.join(
                path,
                "dosdevices/c:",
                "ProgramData/Epic/EpicGamesLauncher/Data/Manifests",
            )
        elif not path.rstrip("/").endswith(
                "ProgramData/Epic/EpicGamesLauncher/Data/Manifests"
        ):
            # lower() might or might not be needed in the check
            return False, path, IndicatorReasonsCommon.WRONG_FORMAT
        if os.path.exists(path):
            return True, path, IndicatorReasonsCommon.VALID
        return False, path, IndicatorReasonsCommon.DIR_NOT_EXISTS

    def egl_path_edit_save_cb(self, path):
        if not path or not os.path.exists(path):
            # This is the same as "--unlink"
            self.core.egl.programdata_path = None
            self.core.lgd.config.remove_option("Legendary", "egl_programdata")
            self.core.lgd.config.remove_option("Legendary", "egl_sync")
            # remove EGL GUIDs from all games, DO NOT remove .egstore folders because that would fuck things up.
            for igame in self.core.get_installed_list():
                igame.egl_guid = ""
                self.core.install_game(igame)
        else:
            self.core.egl.programdata_path = path
            self.core.lgd.config.set("Legendary", "egl_programdata", path)

        self.core.lgd.save_config()

    def egl_path_changed(self, path):
        if self.egl_path_edit.is_valid:
            self.ui.egl_sync_check.setEnabled(bool(path))
        self.ui.egl_sync_check.setCheckState(Qt.Unchecked)
        # self.egl_watcher.removePaths([p for p in self.egl_watcher.directories()])
        # self.egl_watcher.addPaths([path])
        self.update_lists()

    def egl_sync_changed(self, state):
        if state == Qt.Unchecked:
            self.import_list.setEnabled(bool(self.import_list.items))
            self.export_list.setEnabled(bool(self.export_list.items))
            self.core.lgd.config.remove_option("Legendary", "egl_sync")
        else:
            self.core.lgd.config.set("Legendary", "egl_sync", str(True))
            # lk: do import/export here since automatic sync was selected
            self.import_list.mark(Qt.Checked)
            self.export_list.mark(Qt.Checked)
            sync_worker = EGLSyncWorker(self.import_list, self.export_list)
            QThreadPool.globalInstance().start(sync_worker)
            self.import_list.setEnabled(False)
            self.export_list.setEnabled(False)
            # self.update_lists()
        self.core.lgd.save_config()

    def update_lists(self):
        # self.egl_watcher.blockSignals(True)
        have_path = False
        if self.core.egl.programdata_path:
            have_path = os.path.exists(self.core.egl.programdata_path)
            if not have_path and os.path.isdir(os.path.dirname(self.core.egl.programdata_path)):
                # NOTE: This might happen if EGL is installed but no games have been installed through it
                os.mkdir(self.core.egl.programdata_path)
                have_path = os.path.isdir(self.core.egl.programdata_path)
            # NOTE: need to clear known manifests to force refresh
            self.core.egl.manifests.clear()
        self.ui.egl_sync_check_label.setEnabled(have_path)
        self.ui.egl_sync_check.setEnabled(have_path)
        self.import_list.populate(have_path)
        self.import_list.setEnabled(have_path)
        self.export_list.populate(have_path)
        self.export_list.setEnabled(have_path)
        # self.egl_watcher.blockSignals(False)


class EGLSyncListItem(QListWidgetItem):
    def __init__(self, game: Union[EGLManifest,InstalledGame], parent=None):
        super(EGLSyncListItem, self).__init__(parent=parent)
        self.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsUserCheckable)
        self.setCheckState(Qt.Unchecked)
        self.core = RareCore.instance().core()
        self.game = game
        self.setText(self.app_title)

    def is_checked(self) -> bool:
        return self.checkState() == Qt.Checked

    @abstractmethod
    def action(self) -> Union[str, bool]:
        pass

    @property
    def app_name(self):
        return self.game.app_name

    @property
    @abstractmethod
    def app_title(self) -> str:
        pass


class EGLSyncExportItem(EGLSyncListItem):
    def __init__(self, game: InstalledGame, parent=None):
        super(EGLSyncExportItem, self).__init__(game=game, parent=parent)

    def action(self) -> Union[str,bool]:
        error = False
        try:
            self.core.egl_export(self.game.app_name)
        except LgndrException as ret:
            error = ret.message
        return error

    @property
    def app_title(self) -> str:
        return self.game.title


class EGLSyncImportItem(EGLSyncListItem):
    def __init__(self, game: EGLManifest, parent=None):
        super(EGLSyncImportItem, self).__init__(game=game, parent=parent)

    def action(self) -> Union[str,bool]:
        error = False
        try:
            self.core.egl_import(self.game.app_name)
        except LgndrException as ret:
            error = ret.message
        return error

    @property
    def app_title(self) -> str:
        return self.core.get_game(self.game.app_name).app_title


class EGLSyncListGroup(QGroupBox):
    action_errors = pyqtSignal(list)

    def __init__(self, parent=None):
        super(EGLSyncListGroup, self).__init__(parent=parent)
        self.ui = Ui_EGLSyncListGroup()
        self.ui.setupUi(self)
        self.ui.list.setFrameShape(QFrame.NoFrame)
        self.rcore = RareCore.instance()
        self.core = RareCore.instance().core()

        self.ui.list.itemDoubleClicked.connect(
            lambda item: item.setCheckState(Qt.Unchecked)
            if item.checkState() != Qt.Unchecked
            else item.setCheckState(Qt.Checked)
        )
        self.ui.list.itemChanged.connect(self.has_selected)

        self.ui.select_all_button.clicked.connect(lambda: self.mark(Qt.Checked))
        self.ui.select_none_button.clicked.connect(lambda: self.mark(Qt.Unchecked))

        self.ui.action_button.clicked.connect(self.action)

        self.action_errors.connect(self.show_errors)

    def has_selected(self):
        for item in self.items:
            if item.is_checked():
                self.ui.action_button.setEnabled(True)
                return
        self.ui.action_button.setEnabled(False)

    def mark(self, state):
        for item in self.items:
            item.setCheckState(state)

    def populate(self, enabled: bool):
        self.ui.label.setVisible(not enabled or not bool(self.ui.list.count()))
        self.ui.list.setVisible(enabled and bool(self.ui.list.count()))
        self.ui.buttons_widget.setVisible(enabled and bool(self.ui.list.count()))

    @abstractmethod
    def action(self):
        pass

    @pyqtSlot(list)
    @abstractmethod
    def show_errors(self, errors: List):
        pass

    @property
    def items(self) -> Iterable[EGLSyncListItem]:
        # for i in range(self.list.count()):
        #     yield self.list.item(i)
        return [self.ui.list.item(i) for i in range(self.ui.list.count())]


class EGLSyncExportGroup(EGLSyncListGroup):
    def __init__(self, parent=None):
        super(EGLSyncExportGroup, self).__init__(parent=parent)
        self.setTitle(self.tr("Exportable games"))
        self.ui.label.setText(self.tr("No games to export to EGL"))
        self.ui.action_button.setText(self.tr("Export"))

    def populate(self, enabled: bool):
        if enabled:
            self.ui.list.clear()
            for item in self.core.egl_get_exportable():
                try:
                    i = EGLSyncExportItem(item, self.ui.list)
                except AttributeError:
                    logger.error(f"{item.app_name} does not work. Ignoring")
                else:
                    self.ui.list.addItem(i)
        super(EGLSyncExportGroup, self).populate(enabled)

    @pyqtSlot(list)
    def show_errors(self, errors: List):
        QMessageBox.warning(
            self.parent(),
            self.tr("The following errors occurred while exporting."),
            "\n".join(errors),
        )

    def action(self):
        errors: List = []
        for item in self.items:
            if item.is_checked():
                if e := item.action():
                    errors.append(e)
        self.populate(True)
        if errors:
            self.action_errors.emit(errors)


class EGLSyncImportGroup(EGLSyncListGroup):
    def __init__(self, parent=None):
        super(EGLSyncImportGroup, self).__init__(parent=parent)
        self.setTitle(self.tr("Importable games"))
        self.ui.label.setText(self.tr("No games to import from EGL"))
        self.ui.action_button.setText(self.tr("Import"))
        self.list_func = self.core.egl_get_importable

    def populate(self, enabled: bool):
        if enabled:
            self.ui.list.clear()
            for item in self.core.egl_get_importable():
                try:
                    i = EGLSyncImportItem(item, self.ui.list)
                except AttributeError:
                    logger.error(f"{item.app_name} does not work. Ignoring")
                else:
                    self.ui.list.addItem(i)
        super(EGLSyncImportGroup, self).populate(enabled)

    @pyqtSlot(list)
    def show_errors(self, errors: List):
        QMessageBox.warning(
            self.parent(),
            self.tr("The following errors occurred while importing."),
            "\n".join(errors),
        )

    def action(self):
        errors: List = []
        for item in self.items:
            if item.is_checked():
                if e := item.action():
                    errors.append(e)
                else:
                    self.rcore.get_game(item.app_name).set_installed(True)
        self.populate(True)
        if errors:
            self.action_errors.emit(errors)


class EGLSyncWorker(QRunnable):
    def __init__(self, import_list: EGLSyncListGroup, export_list: EGLSyncListGroup):
        super(EGLSyncWorker, self).__init__()
        self.setAutoDelete(True)
        self.import_list = import_list
        self.export_list = export_list

    @pyqtSlot()
    def run(self):
        self.import_list.action()
        self.export_list.action()



from logging import getLogger

from PyQt5.QtCore import QFileSystemWatcher, Qt
from PyQt5.QtWidgets import (
    QGroupBox,
    QHeaderView,
    QVBoxLayout,
    QTableView,
)

from rare.shared import LegendaryCoreSingleton
from .env_vars_model import EnvVarsTableModel

logger = getLogger("EnvVars")


class EnvVars(QGroupBox):
    def __init__(self, parent):
        super(EnvVars, self).__init__(parent=parent)
        self.setTitle(self.tr("Environment variables"))

        self.core = LegendaryCoreSingleton()

        self.table_model = EnvVarsTableModel(self.core)
        self.table_view = QTableView(self)
        self.table_view.setModel(self.table_model)
        self.table_view.verticalHeader().sectionPressed.disconnect()
        self.table_view.horizontalHeader().sectionPressed.disconnect()
        self.table_view.verticalHeader().sectionClicked.connect(self.table_model.removeRow)
        self.table_view.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.table_view.horizontalHeader().setStretchLastSection(True)
        self.table_view.setCornerButtonEnabled(False)

        # FIXME: investigate signaling between widgets
        # We use this function to keep an eye on the config.
        # When the user uses for example the wineprefix settings, we need to update the table.
        # With this function, when the config file changes, we update the table.
        # self.config_file_watcher = QFileSystemWatcher([str(self.core.lgd.config_path)], self)
        # self.config_file_watcher.fileChanged.connect(self.table_model.reset)

        row_height = self.table_view.rowHeight(0)
        self.table_view.setMinimumHeight(row_height * 7)

        layout = QVBoxLayout(self)
        layout.addWidget(self.table_view)

    def keyPressEvent(self, e):
        if e.key() == Qt.Key_Delete or e.key() == Qt.Key_Backspace:
            indexes = self.table_view.selectedIndexes()
            if not len(indexes):
                return
            for idx in indexes:
                if idx.column() == 0:
                    self.table_view.model().removeRow(idx.row())
                elif idx.column() == 1:
                    self.table_view.model().setData(idx, "", Qt.EditRole)
        elif e.key() == Qt.Key_Escape:
            e.ignore()

    def reset_model(self):
        self.table_model.reset()

    def update_game(self, app_name):
        self.table_model.load(app_name)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/games/integrations/import_group.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_ImportGroup(object):
    def setupUi(self, ImportGroup):
        ImportGroup.setObjectName("ImportGroup")
        ImportGroup.resize(651, 218)
        ImportGroup.setWindowTitle("ImportGroup")
        ImportGroup.setWindowFilePath("")
        self.import_layout = QtWidgets.QFormLayout(ImportGroup)
        self.import_layout.setLabelAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.import_layout.setObjectName("import_layout")
        self.path_edit_label = QtWidgets.QLabel(ImportGroup)
        self.path_edit_label.setObjectName("path_edit_label")
        self.import_layout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.path_edit_label)
        self.app_name_label = QtWidgets.QLabel(ImportGroup)
        self.app_name_label.setObjectName("app_name_label")
        self.import_layout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.app_name_label)
        self.import_folder_label = QtWidgets.QLabel(ImportGroup)
        self.import_folder_label.setObjectName("import_folder_label")
        self.import_layout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.import_folder_label)
        self.import_folder_check = QtWidgets.QCheckBox(ImportGroup)
        font = QtGui.QFont()
        font.setItalic(True)
        self.import_folder_check.setFont(font)
        self.import_folder_check.setObjectName("import_folder_check")
        self.import_layout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.import_folder_check)
        self.import_dlcs_label = QtWidgets.QLabel(ImportGroup)
        self.import_dlcs_label.setObjectName("import_dlcs_label")
        self.import_layout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.import_dlcs_label)
        self.import_dlcs_check = QtWidgets.QCheckBox(ImportGroup)
        font = QtGui.QFont()
        font.setItalic(True)
        self.import_dlcs_check.setFont(font)
        self.import_dlcs_check.setObjectName("import_dlcs_check")
        self.import_layout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.import_dlcs_check)
        self.import_force_label = QtWidgets.QLabel(ImportGroup)
        self.import_force_label.setObjectName("import_force_label")
        self.import_layout.setWidget(5, QtWidgets.QFormLayout.LabelRole, self.import_force_label)
        self.import_force_check = QtWidgets.QCheckBox(ImportGroup)
        font = QtGui.QFont()
        font.setItalic(True)
        self.import_force_check.setFont(font)
        self.import_force_check.setObjectName("import_force_check")
        self.import_layout.setWidget(5, QtWidgets.QFormLayout.FieldRole, self.import_force_check)
        self.import_button_label = QtWidgets.QLabel(ImportGroup)
        self.import_button_label.setText("Error")
        self.import_button_label.setObjectName("import_button_label")
        self.import_layout.setWidget(6, QtWidgets.QFormLayout.LabelRole, self.import_button_label)
        self.button_info_layout = QtWidgets.QHBoxLayout()
        self.button_info_layout.setObjectName("button_info_layout")
        self.import_button = QtWidgets.QPushButton(ImportGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.import_button.sizePolicy().hasHeightForWidth())
        self.import_button.setSizePolicy(sizePolicy)
        self.import_button.setObjectName("import_button")
        self.button_info_layout.addWidget(self.import_button)
        self.import_layout.setLayout(6, QtWidgets.QFormLayout.FieldRole, self.button_info_layout)
        self.platform_label = QtWidgets.QLabel(ImportGroup)
        self.platform_label.setObjectName("platform_label")
        self.import_layout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.platform_label)
        self.platform_layout = QtWidgets.QHBoxLayout()
        self.platform_layout.setObjectName("platform_layout")
        self.platform_combo = QtWidgets.QComboBox(ImportGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.platform_combo.sizePolicy().hasHeightForWidth())
        self.platform_combo.setSizePolicy(sizePolicy)
        self.platform_combo.setObjectName("platform_combo")
        self.platform_layout.addWidget(self.platform_combo)
        self.platform_tooltip = QtWidgets.QLabel(ImportGroup)
        font = QtGui.QFont()
        font.setItalic(True)
        self.platform_tooltip.setFont(font)
        self.platform_tooltip.setObjectName("platform_tooltip")
        self.platform_layout.addWidget(self.platform_tooltip)
        self.import_layout.setLayout(2, QtWidgets.QFormLayout.FieldRole, self.platform_layout)

        self.retranslateUi(ImportGroup)

    def retranslateUi(self, ImportGroup):
        _translate = QtCore.QCoreApplication.translate
        ImportGroup.setTitle(_translate("ImportGroup", "Import EGL game from a directory"))
        self.path_edit_label.setText(_translate("ImportGroup", "Installation path"))
        self.app_name_label.setText(_translate("ImportGroup", "Override app name"))
        self.import_folder_label.setText(_translate("ImportGroup", "Import all folders"))
        self.import_folder_check.setText(_translate("ImportGroup", "Scan the installation path for game folders and import them"))
        self.import_dlcs_label.setText(_translate("ImportGroup", "Import DLCs"))
        self.import_dlcs_check.setText(_translate("ImportGroup", "If a game has DLCs, try to import them too"))
        self.import_force_label.setText(_translate("ImportGroup", "Force import"))
        self.import_force_check.setText(_translate("ImportGroup", "Import game despite missing files"))
        self.import_button.setText(_translate("ImportGroup", "Import Game"))
        self.platform_label.setText(_translate("ImportGroup", "Platform"))
        self.platform_tooltip.setText(_translate("ImportGroup", "Select the native platform of the game"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    ImportGroup = QtWidgets.QGroupBox()
    ui = Ui_ImportGroup()
    ui.setupUi(ImportGroup)
    ImportGroup.show()
    sys.exit(app.exec_())



import os
import shutil
import subprocess
from configparser import ConfigParser
from logging import getLogger
from typing import Mapping, Dict, List, Tuple

from rare.lgndr.core import LegendaryCore

logger = getLogger("Wine")


# this is a copied function from legendary.utils.wine_helpers, but registry file can be specified
def read_registry(registry: str, wine_pfx: str) -> ConfigParser:
    accepted = ["system.reg", "user.reg"]
    if registry not in accepted:
        raise RuntimeError(f'Unknown target "{registry}" not in {accepted}')
    reg = ConfigParser(comment_prefixes=(';', '#', '/', 'WINE'), allow_no_value=True,
                       strict=False)
    reg.optionxform = str
    reg.read(os.path.join(wine_pfx, 'system.reg'))
    return reg


def execute(command: List, wine_env: Mapping) -> Tuple[str, str]:
    if os.environ.get("container") == "flatpak":
        flatpak_command = ["flatpak-spawn", "--host"]
        for name, value in wine_env.items():
            flatpak_command.append(f"--env={name}={value}")
        command = flatpak_command + command
    try:
        proc = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            # Use the current environment if we are in flatpak or our own if we are on host
            # In flatpak our environment is passed through `flatpak-spawn` arguments
            env=os.environ.copy() if os.environ.get("container") == "flatpak" else wine_env,
            shell=False,
            text=True,
        )
        res = proc.communicate()
    except (FileNotFoundError, PermissionError) as e:
        res = ("", str(e))
    return res


def resolve_path(wine_exec: str, wine_env: Mapping, path: str) -> str:
    path = path.strip().replace("/", "\\")
    # lk: if path does not exist form
    cmd = [wine_exec, "cmd", "/c", "echo", path]
    # lk: if path exists and needs a case-sensitive interpretation form
    # cmd = [wine_cmd, 'cmd', '/c', f'cd {path} & cd']
    out, err = execute(cmd, wine_env)
    out, err = out.strip(), err.strip()
    if not out:
        logger.error("Failed to resolve wine path due to \"%s\"", err)
        return out
    return out.strip('"')


def query_reg_path(wine_exec: str, wine_env: Mapping, reg_path: str):
    raise NotImplementedError


def query_reg_key(wine_exec: str, wine_env: Mapping, reg_path: str, reg_key) -> str:
    cmd = [wine_exec, "reg", "query", reg_path, "/v", reg_key]
    out, err = execute(cmd, wine_env)
    out, err = out.strip(), err.strip()
    if not out:
        logger.error("Failed to query registry key due to \"%s\"", err)
        return out
    lines = out.split("\n")
    keys: Dict = {}
    for line in lines:
        if line.startswith(" "*4):
            key = [x for x in line.split(" "*4, 3) if bool(x)]
            keys.update({key[0]: key[2]})
    return keys.get(reg_key, "")


def convert_to_windows_path(wine_exec: str, wine_env: Mapping, path: str) -> str:
    raise NotImplementedError


def convert_to_unix_path(wine_exec: str, wine_env: Mapping, path: str) -> str:
    path = path.strip().strip('"')
    cmd = [wine_exec, "winepath.exe", "-u", path]
    out, err = execute(cmd, wine_env)
    out, err = out.strip(), err.strip()
    if not out:
        logger.error("Failed to convert to unix path due to \"%s\"", err)
    return os.path.realpath(out) if (out := out.strip()) else out


def wine(core: LegendaryCore, app_name: str = "default") -> str:
    _wine = core.lgd.config.get(
        app_name, "wine_executable", fallback=core.lgd.config.get(
            "default", "wine_executable", fallback=shutil.which("wine")
        )
    )
    return _wine


def environ(core: LegendaryCore, app_name: str = "default") -> Dict:
    # Get a clean environment if we are in flatpak, this environment will be pass
    # to `flatpak-spawn`, otherwise use the system's.
    _environ = {} if os.environ.get("container") == "flatpak" else os.environ.copy()
    _environ.update(core.get_app_environment(app_name))
    _environ["WINEDEBUG"] = "-all"
    _environ["WINEDLLOVERRIDES"] = "winemenubuilder=d;mscoree=d;mshtml=d;"
    _environ["DISPLAY"] = ""
    return _environ


def prefix(core: LegendaryCore, app_name: str = "default") -> str:
    _prefix = core.lgd.config.get(
        app_name, "wine_prefix", fallback=core.lgd.config.get(
            "default", "wine_prefix", fallback=os.path.expanduser("~/.wine")
        )
    )
    return _prefix if os.path.isdir(_prefix) else ""



from abc import abstractmethod
from dataclasses import dataclass
from datetime import datetime
from enum import IntEnum
from logging import getLogger
from typing import Optional, List, Tuple

from PyQt5.QtCore import QObject, pyqtSignal, QRunnable, QThreadPool, QSettings
from legendary.lfs import eos
from legendary.models.game import SaveGameFile, SaveGameStatus, Game, InstalledGame

from rare.lgndr.core import LegendaryCore
from rare.models.install import UninstallOptionsModel, InstallOptionsModel

logger = getLogger("RareGameBase")


@dataclass
class RareSaveGame:
    file: SaveGameFile
    status: SaveGameStatus = SaveGameStatus.NO_SAVE
    dt_local: Optional[datetime] = None
    dt_remote: Optional[datetime] = None
    description: Optional[str] = ""


class RareGameBase(QObject):

    class State(IntEnum):
        IDLE = 0
        RUNNING = 1
        DOWNLOADING = 2
        VERIFYING = 3
        MOVING = 4
        UNINSTALLING = 5
        SYNCING = 6

    class Signals:
        class Progress(QObject):
            start = pyqtSignal()
            update = pyqtSignal(int)
            finish = pyqtSignal(bool)

        class Widget(QObject):
            update = pyqtSignal()

        class Download(QObject):
            enqueue = pyqtSignal(str)
            dequeue = pyqtSignal(str)

        class Game(QObject):
            install = pyqtSignal(InstallOptionsModel)
            installed = pyqtSignal(str)
            uninstall = pyqtSignal(UninstallOptionsModel)
            uninstalled = pyqtSignal(str)
            launched = pyqtSignal(str)
            finished = pyqtSignal(str)

        def __init__(self):
            super(RareGameBase.Signals, self).__init__()
            self.progress = RareGameBase.Signals.Progress()
            self.widget = RareGameBase.Signals.Widget()
            self.download = RareGameBase.Signals.Download()
            self.game = RareGameBase.Signals.Game()

        def __del__(self):
            self.progress.deleteLater()
            self.widget.deleteLater()
            self.download.deleteLater()
            self.game.deleteLater()

    __slots__ = "igame"

    def __init__(self, legendary_core: LegendaryCore, game: Game):
        super(RareGameBase, self).__init__()
        self.signals = RareGameBase.Signals()
        self.core = legendary_core
        self.game: Game = game
        self._state = RareGameBase.State.IDLE

    def __del__(self):
        del self.signals

    @property
    def state(self) -> 'RareGameBase.State':
        return self._state

    @state.setter
    def state(self, state: 'RareGameBase.State'):
        if state != self._state:
            self._state = state
            self.signals.widget.update.emit()

    @property
    def is_idle(self):
        return self.state == RareGameBase.State.IDLE

    @property
    def app_name(self) -> str:
        return self.game.app_name

    @property
    def app_title(self) -> str:
        return self.game.app_title

    @property
    @abstractmethod
    def is_installed(self) -> bool:
        pass

    @abstractmethod
    def set_installed(self, installed: bool) -> None:
        pass

    @property
    def platforms(self) -> Tuple:
        """!
        @brief Property that holds the platforms a game is available for

        @return Tuple
        """
        return tuple(self.game.asset_infos.keys())

    @property
    def default_platform(self) -> str:
        return self.core.default_platform if self.core.default_platform in self.platforms else "Windows"

    @property
    def is_mac(self) -> bool:
        """!
        @brief Property to report if Game has a mac version

        @return bool
        """
        return "Mac" in self.game.asset_infos.keys()

    @property
    def is_win32(self) -> bool:
        """!
        @brief Property to report if Game is 32bit game

        @return bool
        """
        return "Win32" in self.game.asset_infos.keys()

    @property
    def is_origin(self) -> bool:
        """!
        @brief Property to report if a Game is an Origin game

        Legendary and by extenstion Rare can't launch Origin games directly,
        it just launches the Origin client and thus requires a bit of a special
        handling to let the user know.

        @return bool If the game is an Origin game
        """
        return (
            self.game.metadata.get("customAttributes", {}).get("ThirdPartyManagedApp", {}).get("value") == "Origin"
        )

    @property
    def is_overlay(self):
        return self.app_name == eos.EOSOverlayApp.app_name

    @property
    def version(self) -> str:
        """!
        @brief Reports the currently installed version of the Game

        If InstalledGame reports the currently installed version, which might be
        different from the remote version available from EGS. For not installed Games
        it reports the already known version.

        @return str The current version of the game
        """
        return self.igame.version if self.igame is not None else self.game.app_version(self.default_platform)

    @property
    def install_path(self) -> Optional[str]:
        if self.igame:
            return self.igame.install_path
        return None


class RareGameSlim(RareGameBase):

    def __init__(self, legendary_core: LegendaryCore, game: Game):
        super(RareGameSlim, self).__init__(legendary_core, game)
        # None if origin or not installed
        self.igame: Optional[InstalledGame] = self.core.get_installed_game(game.app_name)
        self.saves: List[RareSaveGame] = []

    @property
    def is_installed(self) -> bool:
        return True

    def set_installed(self, installed: bool) -> None:
        pass

    @property
    def auto_sync_saves(self):
        return self.supports_cloud_saves and QSettings().value(
            f"{self.app_name}/auto_sync_cloud",
            QSettings().value("auto_sync_cloud", False, bool),
            bool
        )

    @property
    def save_path(self) -> Optional[str]:
        if self.igame is not None:
            return self.igame.save_path
        return None

    @property
    def latest_save(self) -> Optional[RareSaveGame]:
        if self.saves:
            saves = sorted(self.saves, key=lambda s: s.file.datetime, reverse=True)
            return saves[0]
        return None

    @property
    def save_game_state(self) -> Tuple[SaveGameStatus, Tuple[Optional[datetime], Optional[datetime]]]:
        if self.saves and self.save_path:
            latest = self.latest_save
            # lk: if the save path wasn't known at startup, dt_local will be None
            # In that case resolve the save again before returning
            latest.status, (latest.dt_local, latest.dt_remote) = self.core.check_savegame_state(
                self.save_path, latest.file
            )
            return latest.status, (latest.dt_local, latest.dt_remote)
        return SaveGameStatus.NO_SAVE, (None, None)

    def upload_saves(self, thread=True):
        status, (dt_local, dt_remote) = self.save_game_state

        def _upload():
            logger.info(f"Uploading save for {self.app_title}")
            self.state = RareGameSlim.State.SYNCING
            self.core.upload_save(self.app_name, self.igame.save_path, dt_local)
            self.state = RareGameSlim.State.IDLE
            self.update_saves()

        if not self.supports_cloud_saves:
            return
        if status == SaveGameStatus.NO_SAVE or not dt_local:
            logger.warning("Can't upload non existing save")
            return
        if self.state == RareGameSlim.State.SYNCING:
            logger.error(f"{self.app_title} is already syncing")
            return

        if thread:
            worker = QRunnable.create(lambda: _upload())
            QThreadPool.globalInstance().start(worker)
        else:
            _upload()

    def download_saves(self, thread=True):
        status, (dt_local, dt_remote) = self.save_game_state

        def _download():
            logger.info(f"Downloading save for {self.app_title}")
            self.state = RareGameSlim.State.SYNCING
            self.core.download_saves(self.app_name, self.latest_save.file.manifest_name, self.save_path)
            self.state = RareGameSlim.State.IDLE
            self.update_saves()

        if not self.supports_cloud_saves:
            return
        if status == SaveGameStatus.NO_SAVE or not dt_remote:
            logger.error("Can't download non existing save")
            return
        if self.state == RareGameSlim.State.SYNCING:
            logger.error(f"{self.app_title} is already syncing")
            return

        if thread:
            worker = QRunnable.create(lambda: _download())
            QThreadPool.globalInstance().start(worker)
        else:
            _download()

    def load_saves(self, saves: List[SaveGameFile]):
        """ Use only in a thread """
        self.saves.clear()
        for save in saves:
            if self.save_path:
                status, (dt_local, dt_remote) = self.core.check_savegame_state(self.save_path, save)
                rsave = RareSaveGame(save, status, dt_local, dt_remote)
            else:
                rsave = RareSaveGame(save, SaveGameStatus.SAME_AGE, dt_local=None, dt_remote=save.datetime)
            self.saves.append(rsave)
        self.signals.widget.update.emit()

    def update_saves(self):
        """ Use only in a thread """
        saves = self.core.get_save_games(self.app_name)
        self.load_saves(saves)

    @property
    def is_save_up_to_date(self):
        status, (_, _) = self.save_game_state
        return (status == SaveGameStatus.SAME_AGE) \
            or (status == SaveGameStatus.NO_SAVE)

    @property
    def raw_save_path(self) -> str:
        if self.game.supports_cloud_saves:
            return self.game.metadata.get("customAttributes", {}).get("CloudSaveFolder", {}).get("value")
        return ""

    @property
    def raw_save_path_mac(self) -> str:
        if self.game.supports_mac_cloud_saves:
            return self.game.metadata.get("customAttributes", {}).get('CloudSaveFolder_MAC', {}).get('value')
        return ""

    @property
    def supports_cloud_saves(self):
        return self.game.supports_cloud_saves or self.game.supports_mac_cloud_saves



from PyQt5.QtCore import QObject, pyqtSignal

from .install import InstallOptionsModel, UninstallOptionsModel


class GlobalSignals:

    # set_main_tab_index = pyqtSignal(int)  # tab index
    # update_gamelist = pyqtSignal(list)

    class ApplicationSignals(QObject):
        # int: exit code
        quit =  pyqtSignal(int)
        # str: app_title
        notify = pyqtSignal(str)
        # none
        prefix_updated = pyqtSignal()
        # none
        overlay_installed = pyqtSignal()
        # none
        update_tray = pyqtSignal()
        # none
        update_statusbar = pyqtSignal()

    class GameSignals(QObject):
        # model
        install = pyqtSignal(InstallOptionsModel)
        # str: app_name
        installed = pyqtSignal(str)
        # model
        uninstall = pyqtSignal(UninstallOptionsModel)
        # str: app_name
        uninstalled = pyqtSignal(str)

    class DownloadSignals(QObject):
        # str: app_name
        enqueue = pyqtSignal(str)
        # str: app_name
        dequeue = pyqtSignal(str)

    class DiscordRPCSignals(QObject):
        # str: app_title
        set_title = pyqtSignal(str)
        # none
        apply_settings = pyqtSignal()

    def __init__(self):
        self.application = GlobalSignals.ApplicationSignals()
        self.game = GlobalSignals.GameSignals()
        self.download = GlobalSignals.DownloadSignals()
        self.discord_rpc = GlobalSignals.DiscordRPCSignals()

    def deleteLater(self):
        self.application.deleteLater()
        del self.application
        self.game.deleteLater()
        del self.game
        self.download.deleteLater()
        del self.download
        self.discord_rpc.deleteLater()
        del self.discord_rpc


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/settings/linux.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_LinuxSettings(object):
    def setupUi(self, LinuxSettings):
        LinuxSettings.setObjectName("LinuxSettings")
        LinuxSettings.resize(394, 84)
        self.linux_layout = QtWidgets.QVBoxLayout(LinuxSettings)
        self.linux_layout.setObjectName("linux_layout")
        self.wine_groupbox = QtWidgets.QGroupBox(LinuxSettings)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.wine_groupbox.sizePolicy().hasHeightForWidth())
        self.wine_groupbox.setSizePolicy(sizePolicy)
        self.wine_groupbox.setObjectName("wine_groupbox")
        self.wine_layout = QtWidgets.QFormLayout(self.wine_groupbox)
        self.wine_layout.setLabelAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.wine_layout.setObjectName("wine_layout")
        self.prefix_label = QtWidgets.QLabel(self.wine_groupbox)
        self.prefix_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.prefix_label.setObjectName("prefix_label")
        self.wine_layout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.prefix_label)
        self.prefix_layout = QtWidgets.QVBoxLayout()
        self.prefix_layout.setObjectName("prefix_layout")
        self.wine_layout.setLayout(0, QtWidgets.QFormLayout.FieldRole, self.prefix_layout)
        self.exec_label = QtWidgets.QLabel(self.wine_groupbox)
        self.exec_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.exec_label.setObjectName("exec_label")
        self.wine_layout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.exec_label)
        self.exec_layout = QtWidgets.QVBoxLayout()
        self.exec_layout.setObjectName("exec_layout")
        self.wine_layout.setLayout(1, QtWidgets.QFormLayout.FieldRole, self.exec_layout)
        self.linux_layout.addWidget(self.wine_groupbox)

        self.retranslateUi(LinuxSettings)

    def retranslateUi(self, LinuxSettings):
        _translate = QtCore.QCoreApplication.translate
        LinuxSettings.setWindowTitle(_translate("LinuxSettings", "LinuxSettings"))
        self.wine_groupbox.setTitle(_translate("LinuxSettings", "Wine Settings"))
        self.prefix_label.setText(_translate("LinuxSettings", "Prefix"))
        self.exec_label.setText(_translate("LinuxSettings", "Executable"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    LinuxSettings = QtWidgets.QWidget()
    ui = Ui_LinuxSettings()
    ui.setupUi(LinuxSettings)
    LinuxSettings.show()
    sys.exit(app.exec_())



import logging
from dataclasses import dataclass
from typing import Protocol, List, Optional, Dict

from rare.lgndr.models.downloading import UIUpdate

logger = logging.getLogger("LgndrMonkeys")


class GetBooleanChoiceProtocol(Protocol):
    def __call__(self, prompt: str, default: bool = ...) -> bool:
        ...


def get_boolean_choice_factory(value: bool = True) -> GetBooleanChoiceProtocol:
    def get_boolean_choice(prompt: str, default: bool = value) -> bool:
        logger.debug("get_boolean_choice: %s, default: %s, choice: %s", prompt, default, value)
        return value
    return get_boolean_choice


class SdlPromptProtocol(Protocol):
    def __call__(self, sdl_data: Dict, title: str) -> List[str]:
        ...


def sdl_prompt_factory(install_tag: Optional[List[str]] = None) -> SdlPromptProtocol:
    def sdl_prompt(sdl_data: Dict, title: str) -> List[str]:
        logger.debug("sdl_prompt: %s", title)
        for key in sdl_data.keys():
            logger.debug("%s: %s %s", key, sdl_data[key]["tags"], sdl_data[key]["name"])
        tags = install_tag if install_tag is not None else [""]
        logger.debug("choice: %s, tags: %s", install_tag, tags)
        return tags
    return sdl_prompt


class VerifyStdoutProtocol(Protocol):
    def __call__(self, a0: int, a1: int, a2: float, a3: float) -> None:
        ...


def verify_stdout_factory(callback: VerifyStdoutProtocol = None) -> VerifyStdoutProtocol:
    def verify_stdout(a0: int, a1: int, a2: float, a3: float) -> None:
        if callback is not None and callable(callback):
            callback(a0, a1, a2, a3)
        else:
            logger.info("Verification progress: %d/%d (%.01f%%) [%.1f MiB/s]", a0, a1, a2, a3)
    return verify_stdout


class UiUpdateProtocol(Protocol):
    def __call__(self, status: UIUpdate) -> None:
        ...


def ui_update_factory(callback: UiUpdateProtocol = None) -> UiUpdateProtocol:
    def ui_update(status: UIUpdate) -> None:
        if callback is not None and callable(callback):
            callback(status)
        else:
            logger.info("Installation progress: %s", status)
    return ui_update


class DLManagerSignals:
    _kill = False
    _update = False

    @property
    def kill(self) -> bool:
        self._update = False
        return self._kill

    @kill.setter
    def kill(self, value: bool) -> None:
        self._update = True
        self._kill = value

    @property
    def update(self) -> bool:
        _update = self._update
        self._update = False
        return _update


@dataclass
class LgndrIndirectStatus:
    success: bool = False
    message: str = ""

    def __len__(self):
        if self.message:
            return 2
        else:
            return 0

    def __bool__(self):
        return self.success

    def __getitem__(self, item):
        if item == 0:
            return self.success
        elif item == 1:
            return self.message
        else:
            raise IndexError

    def __iter__(self):
        return iter((self.success, self.message))

    def __str__(self):
        return self.message


class LgndrIndirectLogger:
    def __init__(
        self, status: LgndrIndirectStatus, logger: logging.Logger = None, level: int = logging.ERROR
    ):
        self.logger = logger
        self.level = level
        self.status = status

    def set_logger(self, logger: logging.Logger):
        self.logger = logger

    def set_level(self, level: int):
        self.level = level

    def _log(self, level: int, msg: str):
        self.status.success = True if level < self.level else False
        self.status.message = msg
        if self.logger:
            self.logger.log(level, msg)

    def debug(self, msg: str):
        self._log(logging.DEBUG, msg)

    def info(self, msg: str):
        self._log(logging.INFO, msg)

    def warning(self, msg: str):
        self._log(logging.WARNING, msg)

    def error(self, msg: str):
        self._log(logging.ERROR, msg)

    def critical(self, msg: str):
        self._log(logging.CRITICAL, msg)

    def fatal(self, msg: str):
        self.critical(msg)



import hashlib
import json
import pickle
import zlib
# from concurrent import futures
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Set
from typing import Tuple, Dict, Union, Type, List, Callable

import requests
from PyQt5.QtCore import (
    Qt,
    pyqtSignal,
    QObject,
    QSize,
    QThreadPool,
    QRunnable,
    QRect,
    QRectF,
)
from PyQt5.QtGui import (
    QPixmap,
    QImage,
    QPainter,
    QPainterPath,
    QBrush,
    QTransform,
    QPen,
)
from PyQt5.QtWidgets import QApplication
from legendary.models.game import Game

from rare.lgndr.core import LegendaryCore
from rare.models.image import ImageSize
from rare.models.signals import GlobalSignals
from rare.utils.paths import image_dir, resources_path, desktop_icon_suffix

# from requests_futures.sessions import FuturesSession

if TYPE_CHECKING:
    pass

logger = getLogger("ImageManager")


class ImageManager(QObject):
    class Worker(QRunnable):
        class Signals(QObject):
            # object: Game
            completed = pyqtSignal(object)

        def __init__(self, func: Callable, updates: List, json_data: Dict, game: Game):
            super(ImageManager.Worker, self).__init__()
            self.signals = ImageManager.Worker.Signals()
            self.setAutoDelete(True)
            self.func = func
            self.updates = updates
            self.json_data = json_data
            self.game = game

        def run(self):
            self.func(self.updates, self.json_data, self.game)
            logger.debug(f"Emitting singal for {self.game.app_name} ({self.game.app_title})")
            self.signals.completed.emit(self.game)

    def __init__(self, signals: GlobalSignals, core: LegendaryCore):
        # lk: the ordering in __img_types matters for the order of fallbacks
        # self.__img_types: Tuple = ("DieselGameBoxTall", "Thumbnail", "DieselGameBoxLogo", "DieselGameBox", "OfferImageTall")
        self.__img_types: Tuple = ("DieselGameBoxTall", "Thumbnail", "DieselGameBoxLogo", "OfferImageTall")
        self.__dl_retries = 1
        self.__worker_app_names: Set[str] = set()
        super(QObject, self).__init__()
        self.signals = signals
        self.core = core

        self.image_dir: Path = image_dir()
        if not self.image_dir.is_dir():
            self.image_dir.mkdir()
            logger.info(f"Created image directory at {self.image_dir}")

        self.device = ImageSize.Preset(1, QApplication.instance().devicePixelRatio())

        self.threadpool = QThreadPool()
        self.threadpool.setMaxThreadCount(6)

    def __img_dir(self, app_name: str) -> Path:
        return self.image_dir.joinpath(app_name)

    def __img_json(self, app_name: str) -> Path:
        return self.__img_dir(app_name).joinpath("image.json")

    def __img_cache(self, app_name: str) -> Path:
        return self.__img_dir(app_name).joinpath("image.cache")

    def __img_color(self, app_name: str) -> Path:
        return self.__img_dir(app_name).joinpath("installed.png")

    def __img_gray(self, app_name: str) -> Path:
        return self.__img_dir(app_name).joinpath("uninstalled.png")

    def __img_desktop_icon(self, app_name: str) -> Path:
        return self.__img_dir(app_name).joinpath(f"icon.{desktop_icon_suffix()}")

    def __prepare_download(self, game: Game, force: bool = False) -> Tuple[List, Dict]:
        if force and self.__img_dir(game.app_name).exists():
            self.__img_color(game.app_name).unlink(missing_ok=True)
            self.__img_gray(game.app_name).unlink(missing_ok=True)
            self.__img_desktop_icon(game.app_name).unlink(missing_ok=True)
        if not self.__img_dir(game.app_name).is_dir():
            self.__img_dir(game.app_name).mkdir()

        # Load image checksums
        if not self.__img_json(game.app_name).is_file():
            json_data: Dict = dict(zip(self.__img_types, [None] * len(self.__img_types)))
        else:
            json_data = json.load(open(self.__img_json(game.app_name), "r"))

        # lk: Find updates or initialize if images are missing.
        # lk: `updates` will be empty for games without images
        # lk: so everything below it is skipped
        updates = []
        if not (
            self.__img_color(game.app_name).is_file()
            and self.__img_gray(game.app_name).is_file()
            and self.__img_desktop_icon(game.app_name).is_file()
        ):
            # lk: fast path for games without images, convert Rare's logo
            if not game.metadata.get("keyImages", []):
                cache_data: Dict = dict(zip(self.__img_types, [None] * len(self.__img_types)))
                cache_data["DieselGameBoxTall"] = open(
                    resources_path.joinpath("images", "cover.png"), "rb"
                ).read()
                # cache_data["DieselGameBoxLogo"] = open(
                #         resources_path.joinpath("images", "Rare_nonsquared.png"), "rb").read()
                self.__convert(game, cache_data)
                json_data["cache"] = None
                json_data["scale"] = ImageSize.Image.pixel_ratio
                json_data["size"] = ImageSize.Image.size.__str__()
                json.dump(json_data, open(self.__img_json(game.app_name), "w"))
            else:
                updates = [image for image in game.metadata["keyImages"] if image["type"] in self.__img_types]
        else:
            for image in game.metadata.get("keyImages", []):
                if image["type"] in self.__img_types:
                    if image["type"] not in json_data.keys() or json_data[image["type"]] != image["md5"]:
                        updates.append(image)

        return updates, json_data

    def __download(self, updates, json_data, game, use_async: bool = False) -> bool:
        # Decompress existing image.cache
        if not self.__img_cache(game.app_name).is_file():
            cache_data = dict(zip(self.__img_types, [None] * len(self.__img_types)))
        else:
            cache_data = self.__decompress(game)

        # lk: filter updates again against the cache now that it is available
        updates = [
            image
            for image in updates
            if cache_data.get(image["type"], None) is None or json_data[image["type"]] != image["md5"]
        ]

        # Download
        # # lk: Keep this here, so I don't have to go looking for it again,
        # # lk: it might be useful in the future.
        # if use_async and len(updates) > 1:
        #     session = FuturesSession(max_workers=len(self.__img_types))
        #     image_requests = []
        #     for image in updates:
        #         logger.info(f"Downloading {image['type']} for {game.app_title}")
        #         json_data[image["type"]] = image["md5"]
        #         payload = {"resize": 1, "w": ImageSize.Image.size.width(), "h": ImageSize.Image.size.height()}
        #         req = session.get(image["url"], params=payload)
        #         req.image_type = image["type"]
        #         image_requests.append(req)
        #     for req in futures.as_completed(image_requests):
        #         cache_data[req.image_type] = req.result().content
        # else:
        for image in updates:
            logger.info(f"Downloading {image['type']} for {game.app_name} ({game.app_title})")
            json_data[image["type"]] = image["md5"]
            payload = {"resize": 1, "w": ImageSize.Image.size.width(), "h": ImageSize.Image.size.height()}
            try:
                # cache_data[image["type"]] = requests.get(image["url"], params=payload).content
                cache_data[image["type"]] = requests.get(image["url"], params=payload, timeout=10).content
            except Exception as e:
                logger.error(e)
                return False

        self.__convert(game, cache_data)
        # lk: don't keep the cache if there is no logo (kept for me)
        # if cache_data["DieselGameBoxLogo"] is not None:
        #     self.__compress(game, cache_data)
        self.__compress(game, cache_data)

        # hash image cache
        try:
            with open(self.__img_cache(game.app_name), "rb") as archive:
                archive_hash = hashlib.md5(archive.read()).hexdigest()
        except FileNotFoundError:
            archive_hash = None

        json_data["cache"] = archive_hash
        json_data["scale"] = ImageSize.Image.pixel_ratio
        json_data["size"] = {"w": ImageSize.Image.size.width(), "h": ImageSize.Image.size.height()}

        # write image.json
        with open(self.__img_json(game.app_name), "w") as file:
            json.dump(json_data, file)

        return bool(updates)

    __icon_overlay: Optional[QPainterPath] = None

    @staticmethod
    def __generate_icon_overlay(rect: QRect) -> QPainterPath:
        if ImageManager.__icon_overlay is not None:
            return ImageManager.__icon_overlay
        rounded_path = QPainterPath()
        margin = 0.1
        rounded_path.addRoundedRect(
            QRectF(
                rect.width() * margin,
                rect.height() * margin,
                rect.width() - (rect.width() * margin * 2),
                rect.height() - (rect.width() * margin * 2)
            ),
            rect.height() * 0.2,
            rect.height() * 0.2,
        )
        ImageManager.__icon_overlay = rounded_path
        return ImageManager.__icon_overlay

    @staticmethod
    def __convert_icon(cover: QImage) -> QImage:
        icon_size = QSize(128, 128)
        icon = QImage(icon_size, QImage.Format_ARGB32_Premultiplied)
        painter = QPainter(icon)
        painter.setRenderHint(QPainter.SmoothPixmapTransform, True)
        painter.setRenderHint(QPainter.Antialiasing, True)
        painter.setCompositionMode(QPainter.CompositionMode_Source)
        painter.fillRect(icon.rect(), Qt.transparent)
        overlay = ImageManager.__generate_icon_overlay(icon.rect())
        brush = QBrush(cover)
        scale = max(icon.width()/cover.width(), icon.height()/cover.height())
        transform = QTransform().scale(scale, scale)
        brush.setTransform(transform)
        painter.fillPath(overlay, brush)
        pen = QPen(Qt.black, 2)
        painter.setPen(pen)
        painter.drawPath(overlay)
        painter.end()
        return icon

    def __convert(self, game, images, force=False) -> None:
        for image in [self.__img_color(game.app_name), self.__img_gray(game.app_name)]:
            if force and image.exists():
                image.unlink(missing_ok=True)

        cover_data = None
        for image_type in self.__img_types:
            if images[image_type] is not None:
                cover_data = images[image_type]
                break

        cover = QImage()
        cover.loadFromData(cover_data)
        cover.convertToFormat(QImage.Format_ARGB32_Premultiplied)
        # lk: Images are not always 4/3, crop them to size
        factor = min(cover.width() // 3, cover.height() // 4)
        rem_w = (cover.width() - factor * 3) // 2
        rem_h = (cover.height() - factor * 4) // 2
        cover = cover.copy(rem_w, rem_h, factor * 3, factor * 4)

        if images["DieselGameBoxLogo"] is not None:
            logo = QImage()
            logo.loadFromData(images["DieselGameBoxLogo"])
            logo.convertToFormat(QImage.Format_ARGB32_Premultiplied)
            if logo.width() > cover.width():
                logo = logo.scaled(cover.width(), cover.height(), Qt.KeepAspectRatio, Qt.SmoothTransformation)
            painter = QPainter(cover)
            painter.drawImage((cover.width() - logo.width()) // 2, cover.height() - logo.height(), logo)
            painter.end()

        cover = cover.scaled(ImageSize.Image.size, Qt.KeepAspectRatio, Qt.SmoothTransformation)

        icon = self.__convert_icon(cover)
        icon.save(str(self.__img_desktop_icon(game.app_name)), format=desktop_icon_suffix().upper())

        # this is not required if we ever want to re-apply the alpha channel
        # cover = cover.convertToFormat(QImage.Format_Indexed8)

        # add the alpha channel back to the cover
        cover = cover.convertToFormat(QImage.Format_ARGB32_Premultiplied)

        cover.save(str(self.__img_color(game.app_name)), format="PNG")
        # quick way to convert to grayscale
        cover = cover.convertToFormat(QImage.Format_Grayscale8)
        # add the alpha channel back to the grayscale cover
        cover = cover.convertToFormat(QImage.Format_ARGB32_Premultiplied)
        cover.save(str(self.__img_gray(game.app_name)), format="PNG")

    def __compress(self, game: Game, data: Dict) -> None:
        archive = open(self.__img_cache(game.app_name), "wb")
        cdata = zlib.compress(pickle.dumps(data), level=-1)
        archive.write(cdata)
        archive.close()

    def __decompress(self, game: Game) -> Dict:
        archive = open(self.__img_cache(game.app_name), "rb")
        try:
            data = zlib.decompress(archive.read())
            data = pickle.loads(data)
        except zlib.error:
            data = dict(zip(self.__img_types, [None] * len(self.__img_types)))
        finally:
            archive.close()
        return data

    def download_image(
        self, game: Game, load_callback: Callable[[], None], priority: int, force: bool = False
    ) -> None:
        if game.app_name in self.__worker_app_names:
            return
        self.__worker_app_names.add(game.app_name)
        updates, json_data = self.__prepare_download(game, force)
        if not updates:
            self.__worker_app_names.remove(game.app_name)
            load_callback()
        else:
            image_worker = ImageManager.Worker(self.__download, updates, json_data, game)
            image_worker.signals.completed.connect(lambda g: self.__worker_app_names.remove(g.app_name))
            image_worker.signals.completed.connect(load_callback)
            self.threadpool.start(image_worker, priority)

    def download_image_blocking(self, game: Game, force: bool = False) -> None:
        updates, json_data = self.__prepare_download(game, force)
        if not updates:
            return
        if updates:
            self.__download(updates, json_data, game, use_async=True)

    def __get_cover(
        self, container: Union[Type[QPixmap], Type[QImage]], app_name: str, color: bool = True
    ) -> Union[QPixmap, QImage]:
        ret = container()
        if not app_name:
            raise RuntimeError("app_name is an empty string")
        if color:
            if self.__img_color(app_name).is_file():
                ret.load(str(self.__img_color(app_name)))
        else:
            if self.__img_gray(app_name).is_file():
                ret.load(str(self.__img_gray(app_name)))
        if not ret.isNull():
            ret.setDevicePixelRatio(ImageSize.Image.pixel_ratio)
            # lk: Scaling happens at painting. It might be inefficient so leave this here as an alternative
            # lk: If this is uncommented, the transformation in ImageWidget should be adjusted also
            ret = ret.scaled(self.device.size, Qt.KeepAspectRatio, Qt.SmoothTransformation)
            ret.setDevicePixelRatio(self.device.pixel_ratio)
        return ret

    def get_pixmap(self, app_name: str, color: bool = True) -> QPixmap:
        """
        Use when the image is to be presented directly on the screen.

        @param app_name: The RareGame object for this game
        @param color: True to load the colored pixmap, False to load the grayscale
        @return: QPixmap
        """
        pixmap: QPixmap = self.__get_cover(QPixmap, app_name, color)
        return pixmap

    def get_image(self, app_name: str, color: bool = True) -> QImage:
        """
        Use when the image has to be manipulated before being rendered.

        @param app_name: The RareGame object for this game
        @param color: True to load the colored image, False to load the grayscale
        @return: QImage
        """
        image: QImage = self.__get_cover(QImage, app_name, color)
        return image



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/downloads/download_widget.ui'
#
# Created by: PyQt5 UI code generator 5.15.8
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_DownloadWidget(object):
    def setupUi(self, DownloadWidget):
        DownloadWidget.setObjectName("DownloadWidget")
        DownloadWidget.resize(490, 90)
        DownloadWidget.setWindowTitle("DownloadWidget")
        self.active_download_layout = QtWidgets.QGridLayout(DownloadWidget)
        self.active_download_layout.setContentsMargins(0, 0, 0, 0)
        self.active_download_layout.setObjectName("active_download_layout")
        self.downloaded = QtWidgets.QLabel(DownloadWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.downloaded.sizePolicy().hasHeightForWidth())
        self.downloaded.setSizePolicy(sizePolicy)
        self.downloaded.setText("downloaded")
        self.downloaded.setObjectName("downloaded")
        self.active_download_layout.addWidget(self.downloaded, 2, 3, 1, 1)
        self.dl_speed_label = QtWidgets.QLabel(DownloadWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.dl_speed_label.setFont(font)
        self.dl_speed_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.dl_speed_label.setObjectName("dl_speed_label")
        self.active_download_layout.addWidget(self.dl_speed_label, 1, 2, 1, 1)
        self.dl_name_label = QtWidgets.QLabel(DownloadWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.dl_name_label.setFont(font)
        self.dl_name_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.dl_name_label.setObjectName("dl_name_label")
        self.active_download_layout.addWidget(self.dl_name_label, 0, 0, 1, 1)
        self.cache_used = QtWidgets.QLabel(DownloadWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cache_used.sizePolicy().hasHeightForWidth())
        self.cache_used.setSizePolicy(sizePolicy)
        self.cache_used.setText("cache_used")
        self.cache_used.setObjectName("cache_used")
        self.active_download_layout.addWidget(self.cache_used, 2, 1, 1, 1)
        self.time_left = QtWidgets.QLabel(DownloadWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.time_left.sizePolicy().hasHeightForWidth())
        self.time_left.setSizePolicy(sizePolicy)
        self.time_left.setText("time_left")
        self.time_left.setObjectName("time_left")
        self.active_download_layout.addWidget(self.time_left, 1, 1, 1, 1)
        self.downloaded_label = QtWidgets.QLabel(DownloadWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.downloaded_label.setFont(font)
        self.downloaded_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.downloaded_label.setObjectName("downloaded_label")
        self.active_download_layout.addWidget(self.downloaded_label, 2, 2, 1, 1)
        self.dl_speed = QtWidgets.QLabel(DownloadWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.dl_speed.sizePolicy().hasHeightForWidth())
        self.dl_speed.setSizePolicy(sizePolicy)
        self.dl_speed.setText("dl_speed")
        self.dl_speed.setObjectName("dl_speed")
        self.active_download_layout.addWidget(self.dl_speed, 1, 3, 1, 1)
        self.cache_used_label = QtWidgets.QLabel(DownloadWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.cache_used_label.setFont(font)
        self.cache_used_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.cache_used_label.setObjectName("cache_used_label")
        self.active_download_layout.addWidget(self.cache_used_label, 2, 0, 1, 1)
        self.dl_name = QtWidgets.QLabel(DownloadWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.dl_name.sizePolicy().hasHeightForWidth())
        self.dl_name.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.dl_name.setFont(font)
        self.dl_name.setText("dl_name")
        self.dl_name.setObjectName("dl_name")
        self.active_download_layout.addWidget(self.dl_name, 0, 1, 1, 3)
        self.time_left_label = QtWidgets.QLabel(DownloadWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.time_left_label.setFont(font)
        self.time_left_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.time_left_label.setObjectName("time_left_label")
        self.active_download_layout.addWidget(self.time_left_label, 1, 0, 1, 1)
        self.kill_button = QtWidgets.QPushButton(DownloadWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.kill_button.sizePolicy().hasHeightForWidth())
        self.kill_button.setSizePolicy(sizePolicy)
        self.kill_button.setObjectName("kill_button")
        self.active_download_layout.addWidget(self.kill_button, 3, 4, 1, 1)
        self.progress_bar = QtWidgets.QProgressBar(DownloadWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.progress_bar.sizePolicy().hasHeightForWidth())
        self.progress_bar.setSizePolicy(sizePolicy)
        self.progress_bar.setProperty("value", 0)
        self.progress_bar.setObjectName("progress_bar")
        self.active_download_layout.addWidget(self.progress_bar, 3, 0, 1, 4)
        self.active_download_layout.setColumnStretch(1, 1)
        self.active_download_layout.setColumnStretch(3, 1)

        self.retranslateUi(DownloadWidget)

    def retranslateUi(self, DownloadWidget):
        _translate = QtCore.QCoreApplication.translate
        self.dl_speed_label.setText(_translate("DownloadWidget", "Download speed:"))
        self.dl_name_label.setText(_translate("DownloadWidget", "Current download:"))
        self.downloaded_label.setText(_translate("DownloadWidget", "Downloaded:"))
        self.cache_used_label.setText(_translate("DownloadWidget", "Cache size:"))
        self.time_left_label.setText(_translate("DownloadWidget", "Time left:"))
        self.kill_button.setText(_translate("DownloadWidget", "Stop Download"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    DownloadWidget = QtWidgets.QWidget()
    ui = Ui_DownloadWidget()
    ui.setupUi(DownloadWidget)
    DownloadWidget.show()
    sys.exit(app.exec_())



"""
Shared controller resources module

Each of the objects in this module should be instantiated ONCE
and only ONCE!
"""

import logging
from argparse import Namespace
from typing import Optional

from rare.lgndr.core import LegendaryCore
from rare.models.signals import GlobalSignals
from .image_manager import ImageManager
from .rare_core import RareCore

logger = logging.getLogger("Shared")


def ArgumentsSingleton() -> Optional[Namespace]:
    return RareCore.instance().args()


def GlobalSignalsSingleton() -> GlobalSignals:
    return RareCore.instance().signals()


def LegendaryCoreSingleton() -> LegendaryCore:
    return RareCore.instance().core()


def ImageManagerSingleton() -> ImageManager:
    return RareCore.instance().image_manager()



from enum import Enum
from logging import getLogger
from typing import List, Dict, Tuple, Any, Callable

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtGui import QIntValidator, QDoubleValidator
from PyQt5.QtWidgets import QGroupBox, QCheckBox, QWidget, QLineEdit, QLabel, QComboBox

from rare.shared import LegendaryCoreSingleton
from rare.ui.components.tabs.settings.widgets.overlay import Ui_OverlaySettings
from rare.utils import config_helper

logger = getLogger("GameOverlays")


class TextInputField(QLineEdit):
    def __init__(self):
        super(TextInputField, self).__init__()
        self.value_changed = self.textChanged
        self.set_value = self.setText
        self.set_default = lambda: self.setText("")

    def get_value(self):
        return self.text()


class ComboBox(QComboBox):
    def __init__(self):
        super(ComboBox, self).__init__()
        self.value_changed = self.currentIndexChanged
        self.get_value = self.currentText
        self.set_value = self.setCurrentText
        self.set_default = lambda: self.setCurrentIndex(0)


class CustomOption:
    input_field: QWidget
    var_name: str

    @classmethod
    def string_input(cls, var_name: str, placeholder: str):
        tmp = cls()
        tmp.input_field = TextInputField()
        tmp.var_name = var_name
        tmp.input_field.setPlaceholderText(placeholder)
        return tmp

    @classmethod
    def number_input(cls, var_name: str, placeholder: Any, is_float: bool = False):
        tmp = cls()
        tmp.input_field = TextInputField()
        tmp.var_name = var_name
        tmp.input_field.setPlaceholderText(str(placeholder))
        if is_float:
            validator = QDoubleValidator()
        else:
            validator = QIntValidator()
        tmp.input_field.setValidator(validator)
        return tmp

    @classmethod
    def select_input(cls, var_name: str, options: List[str]):
        """options: default value in options[0]"""
        tmp = cls()
        tmp.input_field = ComboBox()
        tmp.var_name = var_name
        tmp.input_field.addItems(options)
        return tmp


class ActivationStates(Enum):
    DEFAULT = 0
    HIDDEN = 1
    ACTIVATED = 2


class OverlaySettings(QGroupBox, Ui_OverlaySettings):
    # str: option key
    environ_changed = pyqtSignal(str)
    name: str = "default"
    settings_updatable = True

    def __init__(self, checkboxes_map: List[Tuple[str, str]], value_map: List[Tuple[CustomOption, str]],
                 config_env_var_name: str, no_display_value: str,
                 set_activation_state: Callable[[Enum], None] = lambda x: None):
        super(OverlaySettings, self).__init__()
        self.setupUi(self)
        self.core = LegendaryCoreSingleton()
        self.config_env_var_name = config_env_var_name
        self.no_display_value = no_display_value
        self.set_activation_state = set_activation_state

        self.checkboxes: Dict[str, QCheckBox] = {}

        for i, (var_name, translated_text) in enumerate(checkboxes_map):
            cb = QCheckBox(translated_text)
            self.options_grid.addWidget(cb, i // 4, i % 4)
            self.checkboxes[var_name] = cb
            cb.stateChanged.connect(self.update_settings)

        self.values: Dict[str, QWidget] = {}

        num_rows = len(checkboxes_map) // 4
        for custom_option, translated_text in value_map:
            input_field = custom_option.input_field
            self.options_form.addRow(QLabel(translated_text), input_field)
            self.values[custom_option.var_name] = input_field
            input_field.value_changed.connect(self.update_settings)
            num_rows += 1

        self.show_overlay_combo.currentIndexChanged.connect(self.update_settings)

    def update_settings(self):
        if not self.settings_updatable:
            return
        if self.show_overlay_combo.currentIndex() == 0:
            # System default
            config_helper.remove_option(f"{self.name}.env", self.config_env_var_name)
            self.environ_changed.emit(self.config_env_var_name)
            self.gb_options.setDisabled(True)
            self.set_activation_state(ActivationStates.DEFAULT)
            return

        elif self.show_overlay_combo.currentIndex() == 1:
            # hidden
            config_helper.add_option(f"{self.name}.env", self.config_env_var_name, self.no_display_value)
            self.environ_changed.emit(self.config_env_var_name)
            self.gb_options.setDisabled(True)
            self.set_activation_state(ActivationStates.HIDDEN)
            return
        elif self.show_overlay_combo.currentIndex() == 2:
            self.gb_options.setDisabled(False)
            # custom options
            var_names = []
            for var_name, cb in self.checkboxes.items():
                if cb.isChecked():
                    var_names.append(var_name)

            for var_name, input_field in self.values.items():
                text = input_field.get_value()
                if text not in ["default", ""]:
                    var_names.append(f"{var_name}={text}")

            if not var_names:
                list(self.checkboxes.values())[0].setChecked(True)
                var_names.append(list(self.checkboxes.keys())[0])

            config_helper.add_option(f"{self.name}.env", self.config_env_var_name, ",".join(var_names))
            self.environ_changed.emit(self.config_env_var_name)
            self.set_activation_state(ActivationStates.ACTIVATED)

    def load_settings(self, name: str):
        self.settings_updatable = False
        # load game specific
        self.name = name

        for checkbox in self.checkboxes.values():
            checkbox.setChecked(False)
        for input_field in self.values.values():
            input_field.set_default()

        options = self.core.lgd.config.get(f"{self.name}.env", self.config_env_var_name, fallback=None)
        if options is None:
            logger.debug(f"No Overlay settings found {self.config_env_var_name}")
            self.show_overlay_combo.setCurrentIndex(0)
            self.gb_options.setDisabled(True)

        elif options == self.no_display_value:
            # not visible
            self.gb_options.setDisabled(True)
            self.show_overlay_combo.setCurrentIndex(1)

        else:
            self.show_overlay_combo.setCurrentIndex(2)
            for option in options.split(","):
                try:
                    if "=" in option:
                        var_name, value = option.split("=")
                        if var_name in self.checkboxes.keys():
                            self.checkboxes[var_name].setChecked(False)
                        else:
                            self.values[var_name].set_value(value)
                    else:
                        self.checkboxes[option].setChecked(True)
                except Exception as e:
                    logger.warning(e)

            self.gb_options.setDisabled(False)

        self.settings_updatable = True



import platform
from logging import getLogger

from PyQt5.QtCore import QSettings, Qt
from PyQt5.QtWidgets import (
    QWidget,
    QLabel
)

from rare.components.tabs.settings.widgets.env_vars import EnvVars
from rare.components.tabs.settings.widgets.linux import LinuxSettings
from rare.components.tabs.settings.widgets.proton import ProtonSettings
from rare.components.tabs.settings.widgets.wrapper import WrapperSettings
from rare.shared import LegendaryCoreSingleton
from rare.ui.components.tabs.settings.game_settings import Ui_GameSettings

logger = getLogger("GameSettings")


class DefaultGameSettings(QWidget):
    # variable to no update when changing game
    change = False
    app_name: str

    def __init__(self, is_default, parent=None):
        super(DefaultGameSettings, self).__init__(parent=parent)
        self.ui = Ui_GameSettings()
        self.ui.setupUi(self)
        self.core = LegendaryCoreSingleton()
        self.settings = QSettings()

        self.wrapper_settings = WrapperSettings()

        self.ui.launch_settings_group.layout().addRow(
            QLabel("Wrapper"), self.wrapper_settings
        )

        self.env_vars = EnvVars(self)
        self.ui.game_settings_layout.addWidget(self.env_vars)

        if platform.system() != "Windows":
            self.linux_settings = LinuxAppSettings()
            if platform.system() != "Darwin":
                self.proton_settings = ProtonSettings(self.linux_settings, self.wrapper_settings)
                self.ui.proton_layout.addWidget(self.proton_settings)
                self.proton_settings.environ_changed.connect(self.env_vars.reset_model)

            # FIXME: Remove the spacerItem and margins from the linux settings
            # FIXME: This should be handled differently at soem point in the future
            # NOTE: specerItem has been removed
            self.linux_settings.layout().setContentsMargins(0, 0, 0, 0)
            # FIXME: End of FIXME
            self.ui.linux_settings_layout.addWidget(self.linux_settings)
            self.ui.linux_settings_layout.setAlignment(Qt.AlignTop)

            self.ui.game_settings_layout.setAlignment(Qt.AlignTop)

            self.linux_settings.mangohud.set_wrapper_activated.connect(
                lambda active: self.wrapper_settings.add_wrapper("mangohud")
                if active else self.wrapper_settings.delete_wrapper("mangohud"))
            self.linux_settings.environ_changed.connect(self.env_vars.reset_model)
        else:
            self.ui.linux_settings_widget.setVisible(False)

        if is_default:
            self.ui.launch_settings_layout.removeRow(self.ui.skip_update)
            self.ui.launch_settings_layout.removeRow(self.ui.offline)
            self.ui.launch_settings_layout.removeRow(self.ui.launch_params)

            self.load_settings("default")

    def load_settings(self, app_name):
        self.app_name = app_name
        self.wrapper_settings.load_settings(app_name)
        if platform.system() != "Windows":
            self.linux_settings.update_game(app_name)
            proton = self.wrapper_settings.wrappers.get("proton", "")
            if proton:
                proton = proton.text
            if platform.system() != "Darwin":
                self.proton_settings.load_settings(app_name, proton)
            else:
                proton = ""
            if proton:
                self.linux_settings.ui.wine_groupbox.setEnabled(False)
            else:
                self.linux_settings.ui.wine_groupbox.setEnabled(True)
        self.env_vars.update_game(app_name)


class LinuxAppSettings(LinuxSettings):
    def __init__(self):
        super(LinuxAppSettings, self).__init__()

    def update_game(self, app_name):
        self.name = app_name
        self.wine_prefix.setText(self.load_prefix())
        self.wine_exec.setText(self.load_setting(self.name, "wine_executable"))

        self.dxvk.load_settings(self.name)

        self.mangohud.load_settings(self.name)



from PyQt5.QtCore import Qt, pyqtSignal, QCoreApplication
from PyQt5.QtGui import QCloseEvent, QKeyEvent
from PyQt5.QtWidgets import (
    QDialog,
    QLabel,
    QVBoxLayout,
    QCheckBox,
    QHBoxLayout,
    QPushButton,
)
from legendary.utils.selective_dl import get_sdl_appname

from rare.models.game import RareGame
from rare.models.install import UninstallOptionsModel
from rare.utils.misc import icon


class UninstallDialog(QDialog):
    result_ready = pyqtSignal(UninstallOptionsModel)

    def __init__(self, rgame: RareGame, options: UninstallOptionsModel, parent=None):
        super(UninstallDialog, self).__init__(parent=parent)
        self.setAttribute(Qt.WA_DeleteOnClose, True)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint)
        header = self.tr("Uninstall")
        self.setWindowTitle(f'{header} "{rgame.app_title}" - {QCoreApplication.instance().applicationName()}')
        self.info_text = QLabel(
            self.tr("Do you really want to uninstall <b>{}</b>?").format(rgame.app_title)
        )

        self.keep_files = QCheckBox(self.tr("Keep game files."))
        self.keep_files.setChecked(bool(options.keep_files))
        self.keep_config = QCheckBox(self.tr("Keep game configuation."))
        self.keep_config.setChecked(bool(options.keep_config))

        self.uninstall_button = QPushButton(
            icon("ei.remove-circle", color="red"), self.tr("Uninstall")
        )
        self.uninstall_button.setObjectName("UninstallButton")
        self.uninstall_button.clicked.connect(self.__on_uninstall)

        self.cancel_button = QPushButton(self.tr("Cancel"))
        self.cancel_button.clicked.connect(self.__on_cancel)

        form_layout = QVBoxLayout()
        form_layout.setContentsMargins(-1, -1, 0, -1)
        form_layout.addWidget(self.keep_files)
        form_layout.addWidget(self.keep_config)

        button_layout = QHBoxLayout()
        button_layout.addWidget(self.cancel_button)
        button_layout.addStretch(1)
        button_layout.addWidget(self.uninstall_button)

        layout = QVBoxLayout()
        layout.addWidget(self.info_text)
        layout.addLayout(form_layout)
        layout.addLayout(button_layout)

        self.setLayout(layout)

        if get_sdl_appname(rgame.app_name) is not None:
            self.keep_config.setChecked(True)

        self.options: UninstallOptionsModel = options

    def closeEvent(self, a0: QCloseEvent) -> None:
        self.result_ready.emit(self.options)
        super(UninstallDialog, self).closeEvent(a0)

    def __on_uninstall(self):
        self.options.values = (True, self.keep_files.isChecked(), self.keep_config.isChecked())
        self.close()

    def __on_cancel(self):
        self.options.values = (None, None, None)
        self.close()

    def keyPressEvent(self, e: QKeyEvent) -> None:
        if e.key() == Qt.Key_Escape:
            e.accept()
            self.__on_cancel()
